from __future__ import annotations

from datetime import datetime

import matplotlib.pyplot as plt
import numpy as np
import plotly.graph_objects as go

detailed_plots = [
    "2614_timemap",
    "peak_fits",
    "cal_fit",
    "fwhm_fit",
    "cut_spectrum",
    "survival_frac",
    "spectrum",
    "logged_spectrum",
    "peak_track",
]

aoe_plots = [
    "plot_dt_dep",
    "compt_bands_uncorrected",
    "mean_fit",
    "sigma_fit",
    "compt_bands_corrected",
    "cut_fit",
    "classifier",
    "survival_fractions",
    "spectrum",
    "sf_v_energy",
]

lq_plots = [
    "stability",
    "spectrum",
    "sf_v_energy",
    "survival_fractions",
    "cut_fit",
    "classifier",
    "drift_time",
]

baseline_plots = ["baseline_timemap"]

tau_plots = ["slope", "waveforms"]

optimisation_plots = [
    "trap_kernel",
    "zac_kernel",
    "cusp_kernel",
    "trap_acq",
    "zac_acq",
    "cusp_acq",
]

all_detailed_plots = {
    "cuspEmax_ctc_cal": detailed_plots,
    "zacEmax_ctc_cal": detailed_plots,
    "trapEmax_ctc_cal": detailed_plots,
    "trapTmax_cal": detailed_plots,
    "Baseline": baseline_plots,
    "A/E": aoe_plots,
    "LQ": lq_plots,
    "PZ": tau_plots,
    "Optimisation": optimisation_plots,
}


def plot_spectrum(plot_dict, channel, log=False):
    fig = go.Figure()
    bins = plot_dict["bins"]
    counts = plot_dict["counts"]

    fig.add_trace(
        go.Scatter(x=bins, y=counts, name=channel, line_shape="hvh", line={"width": 1})
    )

    fig.update_traces(mode="lines")

    fig.update_layout(
        xaxis={
            "showline": True,
            "showgrid": True,
            "showticklabels": True,
            "linecolor": "grey",
            "linewidth": 2,
            "ticks": "outside",
            "tickfont": {
                "family": "Arial",
                "size": 12,
                "color": "rgb(82, 82, 82)",
            },
        },
        yaxis={
            "showgrid": True,
            "showline": True,
            "linecolor": "grey",
            "linewidth": 2,
            "showticklabels": True,
            "tickfont": {
                "family": "Arial",
                "size": 12,
                "color": "rgb(82, 82, 82)",
            },
        },
        autosize=False,
        margin={
            "autoexpand": False,
            "l": 100,
            "r": 20,
            "t": 110,
        },
        showlegend=False,
        plot_bgcolor="white",
    )
    annotations = []
    # Title
    annotations.append(
        {
            "xref": "paper",
            "yref": "paper",
            "x": 0.2,
            "y": 1.05,
            "xanchor": "left",
            "yanchor": "bottom",
            "text": channel,
            "font": {
                "family": "Arial",
                "size": 20,
                "color": "rgb(82, 82, 82)",
            },
            "showarrow": False,
        }
    )
    # X label
    annotations.append(
        {
            "xref": "paper",
            "yref": "paper",
            "x": 0.5,
            "y": -0.1,
            "xanchor": "center",
            "yanchor": "top",
            "text": "Energy (keV)",
            "font": {
                "family": "Arial",
                "size": 12,
                "color": "rgb(82, 82, 82)",
            },
            "showarrow": False,
        }
    )

    # Y label
    annotations.append(
        {
            "xref": "paper",
            "yref": "paper",
            "x": -0.1,
            "y": 0.5,
            "xanchor": "left",
            "yanchor": "middle",
            "text": "Counts",
            "textangle": 270,
            "font": {
                "family": "Arial",
                "size": 12,
                "color": "rgb(82, 82, 82)",
            },
            "showarrow": False,
        }
    )

    fig.update_layout(yaxis={"showexponent": "all", "exponentformat": "none"})
    if log is True:
        fig.update_yaxes(
            type="log",
        )
        fig.update_layout(yaxis={"showexponent": "all", "exponentformat": "power"})
    fig.update_layout(annotations=annotations)
    return fig


def plot_survival_frac(plot_dict):
    fig = plt.figure()
    plt.step(plot_dict["bins"], plot_dict["sf"], where="mid")
    plt.xlabel("Energy (keV)")
    plt.ylabel("Survival Fraction (%)")
    plt.close()
    return fig


def plot_cut_spectra(plot_dict):
    fig = plt.figure()
    plt.step(plot_dict["bins"], plot_dict["counts"], where="mid", label="After Cuts")
    plt.step(
        plot_dict["bins"], plot_dict["cut_counts"], where="mid", label="Cut Spectrum"
    )
    if np.isnan(plot_dict["pulser_counts"]).all():
        pass
    else:
        plt.step(
            plot_dict["bins"], plot_dict["pulser_counts"], where="mid", label="Pulser"
        )

    plt.xlabel("Energy (keV)")
    plt.ylabel("Counts")
    plt.yscale("log")
    plt.legend(loc="upper right")
    plt.close()
    return fig


def track_peaks(plot_dict):
    time_bins = plot_dict["2614_stability"]["time"]
    th_counts = plot_dict["2614_stability"]["energy"]
    th_spread = plot_dict["2614_stability"]["spread"]
    th_mean = np.mean(th_counts[~np.isnan(th_counts)][:10])
    th_shift = 100 * (th_counts - th_mean) / th_mean
    th_shift_err = 100 * th_spread / th_mean

    lep_counts = plot_dict["583_stability"]["energy"]
    lep_spread = plot_dict["583_stability"]["spread"]
    lep_mean = np.mean(lep_counts[~np.isnan(lep_counts)][:10])
    lep_shift = 100 * (lep_counts - lep_mean) / lep_mean
    lep_shift_err = 100 * lep_spread / lep_mean

    fig = plt.figure(figsize=(8, 6))
    plt.step(time_bins, th_shift, where="mid", label="2.6 MeV peak", color="blue")

    plt.fill_between(
        time_bins,
        th_shift - th_shift_err,
        th_shift + th_shift_err,
        step="mid",
        alpha=0.1,
        color="blue",
    )

    plt.step(time_bins, lep_shift, where="mid", label="583 keV peak", color="orange")
    plt.fill_between(
        time_bins,
        lep_shift - lep_shift_err,
        lep_shift + lep_shift_err,
        step="mid",
        alpha=0.1,
        color="orange",
    )

    pulser_counts = plot_dict["pulser_stability"]["energy"]
    pulser_spread = plot_dict["pulser_stability"]["spread"]
    if np.isnan(pulser_counts).all():
        pass
    else:
        pulser_mean = np.mean(pulser_counts[~np.isnan(pulser_counts)][:10])
        pulser_shift = 100 * (pulser_counts - pulser_mean) / pulser_mean
        pulser_shift_err = 100 * pulser_spread / pulser_mean
        plt.step(
            time_bins,
            pulser_shift,
            where="mid",
            color="red",
            label=f"{pulser_mean:.0f} keV Pulser",
        )
        plt.fill_between(
            time_bins,
            pulser_shift - pulser_shift_err,
            pulser_shift + pulser_shift_err,
            step="mid",
            alpha=0.1,
            color="red",
        )

    ticks, labels = plt.xticks()
    plt.xlabel(
        f"Time starting : {datetime.utcfromtimestamp(ticks[0]).strftime('%d/%m/%y %H:%M')}"
    )
    plt.ylabel("% Shift")
    plt.xticks(
        ticks,
        [datetime.utcfromtimestamp(tick).strftime("%H:%M") for tick in ticks],
    )
    plt.xlim([time_bins[0] - 10, time_bins[-1] + 10])

    plt.grid(which="both")
    plt.legend()
    plt.ylim([-1, 1])
    plt.close()
    return fig



from __future__ import annotations

from legenddashboard.muon.muon_plots import (
    muon_plot_calshift,
    muon_plot_intlight,
    muon_plot_ratesFloor,
    muon_plot_ratesPillBox,
    muon_plot_ratesWall,
    muon_plot_spectra,
    muon_plot_spp,
    muon_plot_totalRates_daily,
    muon_plot_totalRates_hourly,
)

muon_plots_cal_dict = {
    "Cal. Spectra": muon_plot_spectra,
    "Cal. SPP Sigma": muon_plot_spp,
    "Cal. SPP Shift": muon_plot_calshift,
}


muon_plots_mon_dict = {
    "Integral Light": muon_plot_intlight,
    "Total Rates/H": muon_plot_totalRates_hourly,
    "Total Rates/D": muon_plot_totalRates_daily,
    "Pillbox Rates": muon_plot_ratesPillBox,
    "Floor Rates": muon_plot_ratesFloor,
    "Wall Rates": muon_plot_ratesWall,
}



from __future__ import annotations

from legenddashboard.geds.cal.detailed_plots import (
    all_detailed_plots,
    detailed_plots,
    plot_cut_spectra,
    plot_spectrum,
    plot_survival_frac,
    track_peaks,
)
from legenddashboard.geds.cal.summary_plots import (
    get_aoe_results,
    plot_alpha,
    plot_aoe_status,
    plot_baseline_stability,
    plot_bls,
    plot_counts,
    plot_energy_residuals,
    plot_energy_resolutions_2614,
    plot_energy_resolutions_Qbb,
    plot_energy_spectra,
    plot_fep_stability_channels2d,
    plot_no_fitted_aoe_slices,
    plot_no_fitted_energy_peaks,
    plot_pulser_stability_channels2d,
    plot_pz_consts,
    plot_status,
)
from legenddashboard.geds.cal.tracking_plots import (
    plot_aoe_cut,
    plot_aoe_mean,
    plot_aoe_sig,
    plot_ctc_const,
    plot_energy,
    plot_energy_res_2614,
    plot_energy_res_Qbb,
    plot_energy_residuals_period,
    plot_tau,
    plot_tracking,
)

__all__ = [
    "all_detailed_plots",
    "detailed_plots",
    "plot_cut_spectra",
    "plot_energy_residuals_period",
    "plot_spectrum",
    "plot_survival_frac",
    "plot_tracking",
    "track_peaks",
]

tracking_plots = {
    "Energy Calib. Const.": plot_energy,
    "FWHM Qbb": plot_energy_res_Qbb,
    "FWHM FEP": plot_energy_res_2614,
    "Energy Residuals": plot_energy_residuals_period,
    "A/E Mean": plot_aoe_mean,
    "A/E Cut": plot_aoe_cut,
    "A/E Sigma": plot_aoe_sig,
    "PZ": plot_tau,
    "Alpha": plot_ctc_const,
}

summary_plots = {
    "Detector Status": plot_status,
    "Valid. E": plot_no_fitted_energy_peaks,
    "Energy Spectrum": plot_energy_spectra,
    "FEP Counts": plot_counts,
    "FWHM Qbb": plot_energy_resolutions_Qbb,
    "FWHM FEP": plot_energy_resolutions_2614,
    "Energy Residuals": plot_energy_residuals,
    "A/E Status": plot_aoe_status,
    "Valid. A/E": plot_no_fitted_aoe_slices,
    "A/E SF": get_aoe_results,
    "PZ": plot_pz_consts,
    "CT Alpha": plot_alpha,
    "Baseline Spectrum": plot_bls,
    "Baseline Stability": plot_baseline_stability,
    "FEP Stability": plot_fep_stability_channels2d,
    "Pulser Stability": plot_pulser_stability_channels2d,
}



from __future__ import annotations

import colorcet as cc
import numpy as np
import pandas as pd
from bokeh.models import DatetimeTickFormatter, ZoomInTool, ZoomOutTool
from bokeh.plotting import figure


def sipm_plot_vsTime(
    data_barrel, barrel, resample_unit, name_dict, run, period, run_dict
):
    # add two hours to the x values with if condition
    if data_barrel.index[0].utcoffset() != pd.Timedelta(hours=2):
        data_barrel.index += pd.Timedelta(hours=2)

    p = figure(
        width=1000,
        height=600,
        x_axis_type="datetime",
        tools="pan, box_zoom, ywheel_zoom, hover,reset,save",
        active_scroll="ywheel_zoom",
    )
    p.title.text = (
        f"{run_dict['experiment']}-{period}-{run} | SiPM | Light Intensity | {barrel}"
    )
    p.title.align = "center"
    p.title.text_font_size = "25px"
    p.hover.formatters = {"$x": "datetime", "$y": "printf"}
    p.hover.tooltips = [
        ("Time", "$x{%F %H:%M:%S CET}"),
        ("Light Intensity Rate (PE/s)", "$y"),
        ("Channel", "$name"),
    ]

    p.hover.mode = "vline"

    level = 1
    zoom_in = ZoomInTool(
        level=level, dimensions="height", factor=0.5
    )  # set specific zoom factor
    zoom_out = ZoomOutTool(level=level, dimensions="height", factor=0.5)
    p.add_tools(zoom_in, zoom_out)
    # p.toolbar.active_drag = None      use this line to activate only hover and ywheel_zoom as active tool

    len_colours = len(data_barrel.columns)
    colours = cc.palette["glasbey_category10"][:len_colours]

    if resample_unit == "1min":
        for i, col in enumerate(data_barrel):
            p.line(
                "time",
                col,
                source=data_barrel,
                color=colours[i],
                line_width=2.5,
                legend_label=name_dict[int(col[2:])],
                name=col,
            )
    else:
        data_barrel_resampled = data_barrel.resample(
            resample_unit, origin="start"
        ).mean()
        for i, col in enumerate(data_barrel_resampled):
            p.line(
                "time",
                col,
                source=data_barrel_resampled,
                color=colours[i],
                line_width=2.5,
                legend_label=name_dict[int(col[2:])],
                name=col,
            )

    p.legend.location = "bottom_left"
    p.legend.click_policy = "hide"
    p.xaxis.axis_label = (
        f"Time (CET), starting: {data_barrel.index[0].strftime('%d/%m/%Y %H:%M:%S')}"
    )
    p.xaxis.axis_label_text_font_size = "20px"
    p.xaxis.formatter = DatetimeTickFormatter(days="%Y/%m/%d")
    p.yaxis.axis_label = "Light Intensity Rate (PE/s)"
    p.yaxis.axis_label_text_font_size = "20px"

    return p


def sipm_plot_histogram(
    data_barrel, barrel, resample_unit, name_dict, run, period, run_dict
):
    p = figure(
        width=1000,
        height=600,
        y_axis_type="log",
        x_range=(0, 3),
        tools="pan, box_zoom, ywheel_zoom, hover,reset,save",
        active_scroll="ywheel_zoom",
    )
    p.title.text = (
        f"{run_dict['experiment']}-{period}-{run} | SiPM | Light Intensity | {barrel}"
    )
    p.title.align = "center"
    p.title.text_font_size = "25px"
    p.hover.tooltips = [
        ("Light Intensity Rate (PE/s)", "$x"),
        ("Counts", "$y"),
        ("Channel", "$name"),
    ]

    p.hover.mode = "vline"

    level = 1
    zoom_in = ZoomInTool(
        level=level, dimensions="height", factor=0.5
    )  # set specific zoom factor
    zoom_out = ZoomOutTool(level=level, dimensions="height", factor=0.5)
    p.add_tools(zoom_in, zoom_out)
    # p.toolbar.active_drag = None      use this line to activate only hover and ywheel_zoom as active tool

    len_colours = len(data_barrel.columns)
    colours = cc.palette["glasbey_category10"][:len_colours]

    for i, col in enumerate(data_barrel):
        data_channel = data_barrel[col]
        counts_ch, bins_ch = np.histogram(
            data_channel, bins=300, range=(data_channel.min(), 3)
        )
        bins_ch = (bins_ch[:-1] + bins_ch[1:]) / 2
        p.line(
            bins_ch,
            counts_ch,
            color=colours[i],
            line_width=2.5,
            legend_label=name_dict[int(col[2:])],
            name=col,
        )

    p.legend.location = "bottom_left"
    p.legend.click_policy = "hide"
    p.xaxis.axis_label = "Light Intensity Rate (PE/s)"
    p.xaxis.axis_label_text_font_size = "20px"
    p.yaxis.axis_label = "Counts"
    p.yaxis.axis_label_text_font_size = "20px"

    return p



from __future__ import annotations

import colorsys
import datetime as dtt
import logging
from datetime import timedelta

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
from bokeh.layouts import gridplot
from bokeh.models import DatetimeTickFormatter, Legend, ZoomInTool, ZoomOutTool
from bokeh.plotting import figure

log = logging.getLogger(__name__)

PMT_ID = [
    "101 Pillbox",
    "704 Pillbox",
    "102 Pillbox",
    "705 Pillbox",
    "708 Pillbox",
    "104 Pillbox",
    "709 Pillbox",
    "105 Pillbox",
    "710 Pillbox",
    "707 Pillbox",
    "201 Floor",
    "202 Floor",
    "203 Floor",
    "706 Floor",
    "206 Floor",
    "208 Floor",
    "701 Floor",
    "703 Floor",
    "301 Floor",
    "302 Floor",
    "303 Floor",
    "304 Floor",
    "305 Floor",
    "306 Floor",
    "307 Floor",
    "308 Floor",
    "309 Floor",
    "310 Floor",
    "311 Floor",
    "312 Floor",
    "401 Wall",
    "402 Wall",
    "403 Wall",
    "404 Wall",
    "409 Wall",
    "410 Wall",
    "501 Wall",
    "502 Wall",
    "503 Wall",
    "504 Wall",
    "507 Wall",
    "508 Wall",
    "509 Wall",
    "510 Wall",
    "602 Wall",
    "603 Wall",
    "605 Wall",
    "606 Wall",
    "607 Wall",
    "608 Wall",
    "609 Wall",
    "610 Wall",
    "702 Wall",
    "none",
]
num_colors = 54
chan_num = 55

colors = []
for i in range(num_colors):
    hue = i / num_colors
    r, g, b = colorsys.hsv_to_rgb(hue, 1.0, 1.0)
    color = (int(r * 255), int(g * 255), int(b * 255))
    colors.append(color)


def muon_plot_spectra(arrays_dict, run, period, run_dict, plot_type):
    # Extract the Y data column
    y_data = arrays_dict["histo_values"]

    plots = []

    for i in range(9):
        for x in range(
            round(
                ((chan_num - 1) / 9),
            )
            * (i + 0),
            round(
                ((chan_num - 1) / 9),
            )
            * (i + 1),
        ):
            p = figure(
                width=300,
                height=300,
                x_axis_label="Pulse height [LSBs]",
                y_axis_label="counts",
                y_axis_type="log",
                x_range=(0, 100),
                y_range=(1e-0, 2 * np.amax(y_data[i])),
                title="Channel " + str(x) + " (PMT " + str(PMT_ID[x]) + ")",
            )
            p.title.text_font_size = "9pt"
            p.xaxis.axis_label_text_font_size = "10pt"
            p.title.text = f"{run_dict['experiment']}-{period}-{run} | Mu. | Ch. {x} - PMT {PMT_ID[x]}"
            p.title.align = "center"

            # create a ColumnDataSource for each line
            # source = ColumnDataSource(data=dict(x=x_data[x], y=y_data[x]))
            # step = p.step(
            #     "x",
            #     "y",
            #     source=source,
            #     line_width=2,
            #     line_join="round",
            #     line_cap="round",
            #     line_dash="solid",
            #     line_color=colors[x],
            #     mode="center",
            # )
            plots.append(p)

    # return a grid plot with 9 rows and 6 columns
    return gridplot([[plots[j * 6 + i] for i in range(6)] for j in range(9)])


def muon_plot_spp(arrays_dict, run, period, run_dict, plot_type):
    x_data = arrays_dict["mean_LSB"]
    # Extract the Y data column
    y_data = arrays_dict["mean_sigma"]
    dots = []
    # Create a figure object
    p = figure(
        title="SPP gaussian",
        width=1000,
        height=600,
        tools="pan, box_zoom, ywheel_zoom, hover,reset,save",
        active_scroll="ywheel_zoom",
    )
    p.hover.tooltips = [
        ("Pulse Height (LSB)", "$x"),
        ("σ of SPP", "$y"),  # noqa: RUF001
        ("Channel", "$name"),
    ]
    p.title.text = f"{run_dict['experiment']}-{period}-{run} | Muon | SPP Gaussian"
    p.title.align = "center"
    p.title.text_font_size = "25px"
    p.xaxis.axis_label_text_font_size = "20px"
    p.yaxis.axis_label_text_font_size = "20px"

    level = 1
    zoom_in = ZoomInTool(
        level=level, dimensions="height", factor=0.5
    )  # set specific zoom factor
    zoom_out = ZoomOutTool(level=level, dimensions="height", factor=0.5)
    p.add_tools(zoom_in, zoom_out)
    # p.toolbar.active_drag = None      use this line to activate only hover and ywheel_zoom as active tool

    # Set the x and y axis labels and limits
    p.xaxis.axis_label = "Pulse height (LSB)"
    p.yaxis.axis_label = "σ of SPP"  # noqa: RUF001

    p.x_range.start = 0
    p.x_range.end = 100
    p.y_range.start = 0
    p.y_range.end = 78

    legend_list = []
    for channel in range(53):
        dot = p.scatter(
            x_data[channel],
            y_data[channel],
            color=colors[channel],
            marker="circle",
            size=10,
            name=PMT_ID[channel],
        )
        dots.append(dot)
        legend_list.append((PMT_ID[channel], [dot]))

    legend1 = Legend(items=legend_list[:18], orientation="vertical", location=(10, 15))
    legend2 = Legend(
        items=legend_list[18:36], orientation="vertical", location=(10, 15)
    )
    legend3 = Legend(items=legend_list[36:], orientation="vertical", location=(10, 38))
    p.add_layout(legend1, "right")
    p.add_layout(legend2, "right")
    p.add_layout(legend3, "right")
    legend1.click_policy = (
        "hide"  # hide the corresponding line when the legend item is clicked
    )
    legend2.click_policy = (
        "hide"  # hide the corresponding line when the legend item is clicked
    )
    legend3.click_policy = (
        "hide"  # hide the corresponding line when the legend item is clicked
    )

    return p


def muon_plot_calshift(
    x_data, y_data, run, period, run_dict, plot_type
):  # ueberpruefen
    # add a timedelta of 2 hours to convert from UTC to CET
    two_hours = timedelta(hours=2)
    x_data += two_hours
    p = figure(
        title="Calibration mean shift",
        x_axis_label="Date",
        y_axis_label="Mean Shift [LSB]",
        width=1000,
        height=600,
        tools="pan, box_zoom, ywheel_zoom, hover,reset,save",
        active_scroll="ywheel_zoom",
    )
    p.hover.formatters = {"$x": "datetime"}
    p.hover.tooltips = [
        ("Time", "$x{%F %H:%M:%S CET}"),  # Should now correctly display CET timezone
        ("Mean Shift (LSB)", "$y"),
        ("Channel", "$name"),
    ]
    # ...
    level = 1
    zoom_in = ZoomInTool(
        level=level, dimensions="height", factor=0.5
    )  # set specific zoom factor
    zoom_out = ZoomOutTool(level=level, dimensions="height", factor=0.5)
    p.add_tools(zoom_in, zoom_out)
    # p.toolbar.active_drag = None      use this line to activate only hover and ywheel_zoom as active tool

    p.title.text = (
        f"{run_dict['experiment']}-{period}-{run} | Muon | Calibration Mean Shift"
    )
    p.title.align = "center"
    p.title.text_font_size = "25px"
    p.hover.mode = "vline"
    p.title.align = "center"
    p.title.text_font_size = "25px"
    p.xaxis.axis_label_text_font_size = "20px"
    p.yaxis.axis_label_text_font_size = "20px"
    p.xaxis.axis_label = (
        f"Time (CET), starting: {x_data[0][0].strftime('%d/%m/%Y %H:%M:%S')}"
    )

    p.xaxis.formatter = DatetimeTickFormatter(
        hours=["%Y-%m-%d %H:%M"], days=["%Y-%m-%d"], months=["%Y-%m"], years=["%Y"]
    )

    lines = []
    legend_list = []
    for channel in range(53):
        p.scatter(
            x_data[channel],
            y_data[channel],
            color=colors[channel],
            name=PMT_ID[channel],
        )
        line = p.line(
            x_data[channel],
            y_data[channel],
            color=colors[channel],
            name=PMT_ID[channel],
        )
        lines.append(line)
        legend_list.append((PMT_ID[channel], [line]))

    legend1 = Legend(items=legend_list[:18], orientation="vertical", location=(10, 15))
    legend2 = Legend(
        items=legend_list[18:36], orientation="vertical", location=(10, 15)
    )
    legend3 = Legend(items=legend_list[36:], orientation="vertical", location=(10, 38))
    p.add_layout(legend1, "right")
    p.add_layout(legend2, "right")
    p.add_layout(legend3, "right")
    legend1.click_policy = (
        "hide"  # hide the corresponding line when the legend item is clicked
    )
    legend2.click_policy = (
        "hide"  # hide the corresponding line when the legend item is clicked
    )
    legend3.click_policy = (
        "hide"  # hide the corresponding line when the legend item is clicked
    )

    return p


# monitoring plots


def muon_plot_intlight(arrays_dict, period, run, run_dict):
    x_data = arrays_dict["multiplicity"]
    y_data = arrays_dict["int_light"]
    fig, ax = plt.subplots()
    ax.hist2d(
        x_data,
        y_data,
        bins=(54, 1000),
        range=((0, 55), (0.1, 1000)),
        norm=mpl.colors.LogNorm(),
    )
    ax.set_yscale("log")
    ax.set_xlabel("PMT Multiplicity", size=15)
    ax.set_ylabel("Integral Light p.e.", size=15)
    ax.set_title(
        f"{run_dict['experiment']}-{period}-{run} | Muon | Integral Light p.e. vs. Multiplicity"
    )

    return fig


def muon_plot_totalRates_hourly(arrays_dict, period, run, run_dict):
    x_data = arrays_dict["duration"]
    y_data = arrays_dict["red_rates"]

    p = figure(
        x_range=(0, max(x_data) / 3600),
        width=1000,
        height=600,
        tools="pan, box_zoom, ywheel_zoom, hover,reset,save",
        active_scroll="ywheel_zoom",
    )
    p.hover.tooltips = [("Time (H)", "$x"), ("Rate (Hz)", "$y"), ("Channel", "$name")]
    p.hover.mode = "vline"
    p.title.text = f"{run_dict['experiment']}-{period}-{run} | Muon | Red. Hourly Rates for all PMTs"
    p.title.align = "center"
    p.title.text_font_size = "25px"
    p.xaxis.axis_label = "Time (H)"
    p.xaxis.axis_label_text_font_size = "20px"
    p.yaxis.axis_label = "Rate (Hz)"
    p.yaxis.axis_label_text_font_size = "20px"

    level = 1
    zoom_in = ZoomInTool(
        level=level, dimensions="height", factor=0.5
    )  # set specific zoom factor
    zoom_out = ZoomOutTool(level=level, dimensions="height", factor=0.5)
    p.add_tools(zoom_in, zoom_out)
    # p.toolbar.active_drag = None      use this line to activate only hover and ywheel_zoom as active tool

    steps = []
    legend_list = []
    for chan in range(53):
        # log.debug(np.array(x_data)/3600)
        # step=p.step(np.array(x_data)/3600, np.swapaxes(y_data, 0, 1)[chan], line_color=colors[chan], mode="center", name=PMT_ID[chan])
        step = p.line(
            np.array(x_data) / 3600,
            np.swapaxes(y_data, 0, 1)[chan],
            line_color=colors[chan],
            name=PMT_ID[chan],
        )
        steps.append(step)
        legend_list.append((PMT_ID[chan], [step]))

    legend1 = Legend(items=legend_list[:18], orientation="vertical", location=(10, 15))
    legend2 = Legend(
        items=legend_list[18:36], orientation="vertical", location=(10, 15)
    )
    legend3 = Legend(items=legend_list[36:], orientation="vertical", location=(10, 38))
    p.add_layout(legend1, "right")
    p.add_layout(legend2, "right")
    p.add_layout(legend3, "right")
    legend1.click_policy = (
        "hide"  # hide the corresponding line when the legend item is clicked
    )
    legend2.click_policy = (
        "hide"  # hide the corresponding line when the legend item is clicked
    )
    legend3.click_policy = (
        "hide"  # hide the corresponding line when the legend item is clicked
    )

    return p


def muon_plot_totalRates_daily(arrays_dict, period, run, run_dict):
    x_data = arrays_dict["times"]
    y_data = arrays_dict["red_rates"]

    # add a timedelta of two hours to the first entry which is the starting time
    x_label = x_data[0] + timedelta(hours=2)

    p = figure(
        x_axis_type="datetime",
        width=1000,
        height=600,
        tools="pan, box_zoom, ywheel_zoom, hover,reset,save",
        active_scroll="ywheel_zoom",
    )
    p.hover.formatters = {"$x": "datetime"}
    p.hover.tooltips = [
        ("Time", "$x{%F %H:%M:%S CET}"),  # Use the formatted CET
        ("Rate (Hz)", "$snap_y"),
        ("Channel", "$name"),
    ]
    p.hover.mode = "vline"
    p.title.text = f"{run_dict['experiment']}-{period}-{run} | Muon | Red. Daily Rates for all PMTs"
    p.title.align = "center"
    p.title.text_font_size = "25px"
    p.xaxis.axis_label = (
        f"Time (CET), starting: {x_label.strftime('%d/%m/%Y %H:%M:%S')}"
    )
    p.xaxis.axis_label_text_font_size = "20px"
    p.yaxis.axis_label = "Rate (Hz)"
    p.yaxis.axis_label_text_font_size = "20px"
    p.xaxis.formatter = DatetimeTickFormatter(days="%Y/%m/%d")

    level = 1
    zoom_in = ZoomInTool(
        level=level, dimensions="height", factor=0.5
    )  # set specific zoom factor
    zoom_out = ZoomOutTool(level=level, dimensions="height", factor=0.5)
    p.add_tools(zoom_in, zoom_out)
    # p.toolbar.active_drag = None      use this line to activate only hover and ywheel_zoom as active tool

    legend_list = []

    for chan in range(53):
        daily_rates = []
        daily_mean_rates = []
        sum_rate = 0
        count = 0
        current_day = x_data[0].date()

        if len(x_data) != len(np.swapaxes(y_data, 0, 1)[chan]):
            log.debug(
                "lengths of x_data and y_data are not equal",
                extra={
                    "len_ydata": len(np.swapaxes(y_data, 0, 1)[chan]),
                    "len_xdata": len(x_data),
                },
            )
            log.debug("x_data: ", extra={"xdata": x_data})
            log.debug("y_data: ", extra={"ydata": np.swapaxes(y_data, 0, 1)[chan]})
            continue

        for i in range(len(x_data)):
            if x_data[i].date() == current_day:
                sum_rate += np.swapaxes(y_data, 0, 1)[chan][i]
                count += 1
            else:
                daily_rates.append(current_day)
                daily_mean_rates.append(sum_rate / count)
                current_day = x_data[i].date()
                sum_rate = np.swapaxes(y_data, 0, 1)[chan][i]
                count = 1

        daily_rates.append(current_day)
        daily_mean_rates.append(sum_rate / count)
        daily_rates = [
            dtt.datetime.combine(date, dtt.datetime.min.time()) for date in daily_rates
        ]

        # define new array with CET times
        daily_rates_cet = [
            dtt.datetime.combine(date, dtt.datetime.min.time()) for date in daily_rates
        ]
        daily_rates_cet = [dt + timedelta(hours=2) for dt in daily_rates_cet]

        step = p.line(
            daily_rates_cet,
            daily_mean_rates,
            line_color=colors[chan],
            name=PMT_ID[chan],
        )
        legend_list.append((PMT_ID[chan], [step]))

    legend1 = Legend(items=legend_list[:18], orientation="vertical", location=(10, 15))
    legend2 = Legend(
        items=legend_list[18:36], orientation="vertical", location=(10, 15)
    )
    legend3 = Legend(items=legend_list[36:], orientation="vertical", location=(10, 38))
    p.add_layout(legend1, "right")
    p.add_layout(legend2, "right")
    p.add_layout(legend3, "right")
    legend1.click_policy = (
        "hide"  # hide the corresponding line when the legend item is clicked
    )
    legend2.click_policy = (
        "hide"  # hide the corresponding line when the legend item is clicked
    )
    legend3.click_policy = (
        "hide"  # hide the corresponding line when the legend item is clicked
    )

    return p


def muon_plot_ratesPillBox(arrays_dict, period, run, run_dict):
    x_data = arrays_dict["times"]
    y_data = arrays_dict["red_rates"]

    # same steps as in the function above
    x_label = x_data[0] + timedelta(hours=2)

    p = figure(
        x_axis_type="datetime",
        width=1000,
        height=600,
        tools="pan, box_zoom, ywheel_zoom, hover,reset,save",
        active_scroll="ywheel_zoom",
    )
    p.hover.formatters = {"$x": "datetime"}
    p.hover.tooltips = [
        ("Time", "$x{%F %H:%M:%S CET}"),  # Use the formatted CET
        ("Rate (Hz)", "$y"),
        ("Channel", "$name"),
    ]
    p.hover.mode = "vline"
    p.title.text = f"{run_dict['experiment']}-{period}-{run} | Muon | Red. Daily Rates over Time for Pillbox PMTs"
    p.title.align = "center"
    p.title.text_font_size = "25px"
    p.xaxis.axis_label = (
        f"Time (CET), starting: {x_label.strftime('%d/%m/%Y %H:%M:%S')}"
    )
    p.xaxis.axis_label_text_font_size = "20px"
    p.yaxis.axis_label = "Rate (Hz)"
    p.yaxis.axis_label_text_font_size = "20px"
    p.xaxis.formatter = DatetimeTickFormatter(days="%Y/%m/%d")

    level = 1
    zoom_in = ZoomInTool(
        level=level, dimensions="height", factor=0.5
    )  # set specific zoom factor
    zoom_out = ZoomOutTool(level=level, dimensions="height", factor=0.5)
    p.add_tools(zoom_in, zoom_out)
    # p.toolbar.active_drag = None      use this line to activate only hover and ywheel_zoom as active tool

    renderers = {}
    for chan in range(10):
        daily_rates = []
        daily_mean_rates = []
        sum_rate = 0
        count = 0
        current_day = x_data[0].date()

        if len(x_data) != len(np.swapaxes(y_data, 0, 1)[chan]):
            log.debug(
                "lengths of x_data and y_data are not equal",
                extra={
                    "len_ydata": len(np.swapaxes(y_data, 0, 1)[chan]),
                    "len_xdata": len(x_data),
                },
            )
            log.debug("x_data: ", extra={"xdata": x_data})
            log.debug("y_data: ", extra={"ydata": np.swapaxes(y_data, 0, 1)[chan]})
            continue

        for i in range(len(x_data)):
            if x_data[i].date() == current_day:
                sum_rate += np.swapaxes(y_data, 0, 1)[chan][i]
                count += 1
            else:
                daily_rates.append(current_day)
                daily_mean_rates.append(sum_rate / count)
                current_day = x_data[i].date()
                sum_rate = np.swapaxes(y_data, 0, 1)[chan][i]
                count = 1

        daily_rates.append(current_day)
        daily_mean_rates.append(sum_rate / count)
        daily_rates = [
            dtt.datetime.combine(date, dtt.datetime.min.time()) for date in daily_rates
        ]

        daily_rates_cet = [
            dtt.datetime.combine(date, dtt.datetime.min.time()) for date in daily_rates
        ]
        daily_rates_cet = [dt + timedelta(hours=2) for dt in daily_rates]

        # Add a glyph renderer for the line
        # step = p.step(daily_rates, daily_mean_rates, line_color=colors[chan], mode="center", name=PMT_ID[chan])
        step = p.line(
            daily_rates_cet,
            daily_mean_rates,
            line_color=colors[chan],
            name=PMT_ID[chan],
        )
        renderers[PMT_ID[chan]] = step

    # Create a legend with interactive checkboxes
    legend = Legend(
        items=[(label, [renderer]) for label, renderer in renderers.items()],
        location="top_left",
    )
    legend.click_policy = (
        "hide"  # hide the corresponding line when the legend item is clicked
    )
    p.add_layout(legend, "right")

    return p


def muon_plot_ratesFloor(arrays_dict, period, run, run_dict):
    x_data = arrays_dict["times"]
    y_data = arrays_dict["red_rates"]

    x_label = x_data[0] + timedelta(hours=2)

    p = figure(
        x_axis_type="datetime",
        width=1000,
        height=600,
        tools="pan, box_zoom, ywheel_zoom, hover,reset,save",
        active_scroll="ywheel_zoom",
    )
    p.hover.formatters = {"$x": "datetime"}
    p.hover.tooltips = [
        ("Time", "$x{%F %H:%M:%S CET}"),  # Use the formatted CET
        ("Rate (Hz)", "$y"),
        ("Channel", "$name"),
    ]
    p.hover.mode = "vline"
    p.title.text = f"{run_dict['experiment']}-{period}-{run} | Muon | Red. Daily Rates over Time for Floor PMTs"

    level = 1
    zoom_in = ZoomInTool(
        level=level, dimensions="height", factor=0.5
    )  # set specific zoom factor
    zoom_out = ZoomOutTool(level=level, dimensions="height", factor=0.5)
    p.add_tools(zoom_in, zoom_out)
    # p.toolbar.active_drag = None      use this line to activate only hover and ywheel_zoom as active tool

    p.title.align = "center"
    p.title.text_font_size = "25px"
    p.xaxis.axis_label = (
        f"Time (CET), starting: {x_label.strftime('%d/%m/%Y %H:%M:%S')}"
    )
    p.xaxis.axis_label_text_font_size = "20px"
    p.yaxis.axis_label = "Rate (Hz)"
    p.yaxis.axis_label_text_font_size = "20px"
    p.xaxis.formatter = DatetimeTickFormatter(days="%Y/%m/%d")

    renderers = {}
    for chan in range(10, 30):
        daily_rates = []
        daily_mean_rates = []
        sum_rate = 0
        count = 0
        current_day = x_data[0].date()

        if len(x_data) != len(np.swapaxes(y_data, 0, 1)[chan]):
            log.debug(
                "lengths of x_data and y_data are not equal",
                extra={
                    "len_ydata": len(np.swapaxes(y_data, 0, 1)[chan]),
                    "len_xdata": len(x_data),
                },
            )
            log.debug("x_data: ", extra={"xdata": x_data})
            log.debug("y_data: ", extra={"ydata": np.swapaxes(y_data, 0, 1)[chan]})
            continue

        for i in range(len(x_data)):
            if x_data[i].date() == current_day:
                sum_rate += np.swapaxes(y_data, 0, 1)[chan][i]
                count += 1
            else:
                daily_rates.append(current_day)
                daily_mean_rates.append(sum_rate / count)
                current_day = x_data[i].date()
                sum_rate = np.swapaxes(y_data, 0, 1)[chan][i]
                count = 1

        daily_rates.append(current_day)
        daily_mean_rates.append(sum_rate / count)
        daily_rates = [
            dtt.datetime.combine(date, dtt.datetime.min.time()) for date in daily_rates
        ]

        daily_rates_cet = [
            dtt.datetime.combine(date, dtt.datetime.min.time()) for date in daily_rates
        ]
        daily_rates_cet = [dt + timedelta(hours=2) for dt in daily_rates_cet]

        # Add a glyph renderer for the line
        # step = p.step(daily_rates, daily_mean_rates, line_color=colors[chan], mode="center", name=PMT_ID[chan])
        step = p.line(
            daily_rates_cet,
            daily_mean_rates,
            line_color=colors[chan],
            name=PMT_ID[chan],
        )
        renderers[PMT_ID[chan]] = step

    # Create a legend with interactive checkboxes
    legend = Legend(
        items=[(label, [renderer]) for label, renderer in renderers.items()],
        location="top_left",
    )
    legend.click_policy = (
        "hide"  # hide the corresponding line when the legend item is clicked
    )
    p.add_layout(legend, "right")

    return p


def muon_plot_ratesWall(arrays_dict, period, run, run_dict):
    x_data = arrays_dict["times"]
    y_data = arrays_dict["red_rates"]

    x_label = x_data[0] + timedelta(hours=2)

    p = figure(
        x_axis_type="datetime",
        width=1000,
        height=600,
        tools="pan, box_zoom, ywheel_zoom, hover,reset,save",
        active_scroll="ywheel_zoom",
    )
    p.hover.formatters = {"$x": "datetime"}
    p.hover.tooltips = [
        ("Time", "$x{%F %H:%M:%S CET}"),
        ("Rate (Hz)", "$y"),
        ("Channel", "$name"),
    ]
    p.hover.mode = "vline"

    level = 1
    zoom_in = ZoomInTool(
        level=level, dimensions="height", factor=0.5
    )  # set specific zoom factor
    zoom_out = ZoomOutTool(level=level, dimensions="height", factor=0.5)
    p.add_tools(zoom_in, zoom_out)
    # p.toolbar.active_drag = None      use this line to activate only hover and ywheel_zoom as active tool

    p.title.text = f"{run_dict['experiment']}-{period}-{run} | Muon | Red. Daily Rates over Time for Wall PMTs"
    p.title.align = "center"
    p.title.text_font_size = "25px"
    p.xaxis.axis_label = (
        f"Time (CET), starting: {x_label.strftime('%d/%m/%Y %H:%M:%S')}"
    )
    p.xaxis.axis_label_text_font_size = "20px"
    p.yaxis.axis_label = "Rate (Hz)"
    p.yaxis.axis_label_text_font_size = "20px"
    p.xaxis.formatter = DatetimeTickFormatter(days="%Y/%m/%d")

    renderers = {}
    for chan in range(30, 53):
        daily_rates = []
        daily_mean_rates = []
        sum_rate = 0
        count = 0
        current_day = x_data[0].date()

        if len(x_data) != len(np.swapaxes(y_data, 0, 1)[chan]):
            log.debug(
                "lengths of x_data and y_data are not equal",
                extra={
                    "len_ydata": len(np.swapaxes(y_data, 0, 1)[chan]),
                    "len_xdata": len(x_data),
                },
            )
            log.debug("x_data: ", extra={"xdata": x_data})
            log.debug("y_data: ", extra={"ydata": np.swapaxes(y_data, 0, 1)[chan]})
            continue

        for i in range(len(x_data)):
            if x_data[i].date() == current_day:
                sum_rate += np.swapaxes(y_data, 0, 1)[chan][i]
                count += 1
            else:
                daily_rates.append(current_day)
                daily_mean_rates.append(sum_rate / count)
                current_day = x_data[i].date()
                sum_rate = np.swapaxes(y_data, 0, 1)[chan][i]
                count = 1

        daily_rates.append(current_day)
        daily_mean_rates.append(sum_rate / count)
        daily_rates = [
            dtt.datetime.combine(date, dtt.datetime.min.time()) for date in daily_rates
        ]

        daily_rates_cet = [
            dtt.datetime.combine(date, dtt.datetime.min.time()) for date in daily_rates
        ]
        daily_rates_cet = [dt + timedelta(hours=2) for dt in daily_rates_cet]

        # Add a glyph renderer for the line
        # step = p.step(daily_rates, daily_mean_rates, line_color=colors[chan], mode="center", name=PMT_ID[chan])
        step = p.line(
            daily_rates_cet,
            daily_mean_rates,
            line_color=colors[chan],
            name=PMT_ID[chan],
        )
        renderers[PMT_ID[chan]] = step

    # Create a legend with interactive checkboxes
    legend = Legend(
        items=[(label, [renderer]) for label, renderer in renderers.items()],
        location="top_left",
    )
    legend.click_policy = (
        "hide"  # hide the corresponding line when the legend item is clicked
    )
    p.add_layout(legend, "right")

    return p



from __future__ import annotations

import argparse
import datetime as dtt
import logging
import shelve
import time
from pathlib import Path

import numpy as np
import panel as pn
import param
from bokeh.models.widgets.tables import BooleanFormatter
from bokeh.plotting import figure

from legenddashboard import muon
from legenddashboard.base import Monitoring
from legenddashboard.util import logo_path, read_config

log = logging.getLogger(__name__)


class MuonMonitoring(Monitoring):
    muon_path = param.String("")

    # muon plots
    muon_plots_cal = param.ObjectSelector(
        default=next(iter(muon.muon_plots_cal_dict)),
        objects=list(muon.muon_plots_cal_dict),
    )

    muon_plots_mon = param.ObjectSelector(
        default=next(iter(muon.muon_plots_mon_dict)),
        objects=list(muon.muon_plots_mon_dict),
    )
    data_dict = param.Dict({})

    @param.depends("run", watch=True)
    def _get_muon_data(self):
        start_time = time.time()
        data_file = f"{self.muon_path}/generated/plt/phy/{self.period}/dsp/{self.run}/dashboard_period_{self.period}_run_{self.run}.shelve"
        if not (Path(data_file) / ".dat").exists():
            self.muon_data_dict = {}
        else:
            with shelve.open(data_file, "r") as f:
                # Create an empty dictionary
                arrays_dict = {}

                for key in f:
                    # Add a new key-value pair to the dictionary
                    arrays_dict[key] = np.array(f[key])

                self.muon_data_dict = arrays_dict
        log.debug("Time to get muon data:", extra={"time": time.time() - start_time})

    @param.depends("run", "muon_plots_cal")
    def view_muon_cal(self):
        start_time = time.time()
        if not bool(self.muon_data_dict):
            p = figure(width=1000, height=600)
            p.title.text = f"No data for run {self.run_dict[self.run]['experiment']}-{self.period}-{self.run}"
            p.title.align = "center"
            p.title.text_font_size = "25px"
            log.debug(
                "Time to get muon cal plot:", extra={"time": time.time() - start_time}
            )
            return p

        if self.muon_plots_cal == "Cal. SPP Shift":
            data_file = f"{self.muon_path}/generated/plt/phy/{self.period}/dsp/{self.run}/dashboard_period_{self.period}_run_{self.run}.shelve"
            with shelve.open(data_file, "r") as f:
                # x_data_str = np.array(list(f['date'].values()))
                x_data_str = np.array(list(f["date"].values()))
                # y_data = np.array(list(f['mean_shift'].values()))
                y_data = np.array(list(f["mean_shift"].values()))

                # Reshape the x_data and y_data arrays
                x_data = np.array(
                    [
                        [
                            dtt.datetime.strptime(date_str, "%Y_%m_%d")
                            for date_str in row
                        ]
                        for row in x_data_str
                    ]
                )

                p = self.muon_plots_cal_dict[self.muon_plots_cal](
                    x_data,
                    y_data,
                    self.run,
                    self.period,
                    self.run_dict[self.run],
                    self.muon_plots_cal,
                )
                log.debug(
                    "Time to get muon cal plot:",
                    extra={"time": time.time() - start_time},
                )
                return p
        else:
            p = self.muon_plots_cal_dict[self.muon_plots_cal](
                self.muon_data_dict,
                self.run,
                self.period,
                self.run_dict[self.run],
                self.muon_plots_cal,
            )
            log.debug(
                "Time to get muon cal plot:", extra={"time": time.time() - start_time}
            )
            return p

    @param.depends("run", "muon_plots_mon")
    def view_muon_mon(self):
        start_time = time.time()
        if not bool(self.muon_data_dict):
            p = figure(width=1000, height=600)
            p.title.text = f"No data for run {self.run_dict[self.run]['experiment']}-{self.period}-{self.run}"
            p.title.align = "center"
            p.title.text_font_size = "25px"
            log.debug(
                "Time to get muon mon plot:", extra={"time": time.time() - start_time}
            )
            return p
        if self.muon_plots_mon == "Integral Light":
            p = pn.pane.Matplotlib(
                self.muon_plots_mon_dict[self.muon_plots_mon](
                    self.muon_data_dict, self.period, self.run, self.run_dict[self.run]
                ),
                sizing_mode="scale_width",
            )
            log.debug(
                "Time to get muon mon plot:", extra={"time": time.time() - start_time}
            )
            return p
        p = self.muon_plots_mon_dict[self.muon_plots_mon](
            self.muon_data_dict, self.period, self.run, self.run_dict[self.run]
        )
        log.debug(
            "Time to get muon mon plot:", extra={"time": time.time() - start_time}
        )
        return p

    @param.depends("run")
    def view_meta(self):
        start_time = time.time()
        ret = pn.widgets.Tabulator(
            self.meta_df,
            formatters={"Proc.": BooleanFormatter(), "Usabl.": BooleanFormatter()},
            frozen_columns=[0],
        )
        log.debug("Time to get meta:", extra={"time": time.time() - start_time})
        return ret

    def build_muon_cal_pane(self, widget_widths=140):
        muon_cal_param_currentValue = pn.pane.Markdown(f"## {self.muon_plots_cal}")
        muon_cal_param = pn.widgets.MenuButton(
            name="Calibration",
            button_type="primary",
            width=widget_widths,
            items=self.param.muon_plots_cal.objects,
        )

        def update_muon_cal_plots(event):
            self.muon_plots_cal = event.new
            muon_cal_param_currentValue.object = f"## {event.new}"

        muon_cal_param.on_click(update_muon_cal_plots)

        return pn.Column(
            pn.Row(
                pn.pane.SVG(
                    logo_path / "Muon.svg",
                    height=25,
                ),
                muon_cal_param,
            ),
            pn.Row("## Current Plot:", muon_cal_param_currentValue),
            pn.Row(self.view_muon_cal, sizing_mode="scale_both"),
            name="Muon Cal. Plots",
            sizing_mode="scale_both",
        )

    def build_muon_mon_pane(self, widget_widths=140):
        muon_mon_param_currentValue = pn.pane.Markdown(f"## {self.muon_plots_mon}")
        muon_mon_param = pn.widgets.MenuButton(
            name="Monitoring",
            button_type="primary",
            width=widget_widths,
            items=self.param.muon_plots_mon.objects,
        )

        def update_muon_mon_plots(event):
            self.muon_plots_mon = event.new
            muon_mon_param_currentValue.object = f"## {event.new}"

        muon_mon_param.on_click(update_muon_mon_plots)

        return pn.Column(
            pn.Row(
                pn.pane.SVG(
                    logo_path / "Muon.svg",
                    height=25,
                ),
                muon_mon_param,
            ),
            pn.Row("## Current Plot:", muon_mon_param_currentValue),
            pn.Row(self.view_muon_mon, sizing_mode="scale_both"),
            name="Muon Mon. Plots",
            sizing_mode="scale_both",
        )

    def build_muon_panes(
        self,
        widget_widths=140,
    ):
        return {
            "Muon Cal. Plots": self.build_muon_cal_pane(widget_widths),
            "Muon Mon. Plots": self.build_muon_mon_pane(widget_widths),
        }

    @classmethod
    def init_muon_panes(
        cls,
        base_path,
        muon_path,
        widget_widths=140,
    ):
        """
        Initialize the MuonMonitoring class and create the muon panes.

        Args:
            base_path (Path): The base path for the monitoring.
            muon_path (Path): The path to the muon data.

        Returns:
            dict: A dictionary containing the muon panes.
        """
        return cls(
            base_path=base_path,
            muon_path=muon_path,
        ).build_muon_panes(base_path, widget_widths)

    @classmethod
    def display_muon_panes(
        cls,
        base_path,
        notebook=False,
        widget_widths: int = 140,
    ):
        """
        View the calibration panes.

        Args:
            widget_widths (int): Width of the widgets.

        Returns:
            pn.Row: Row containing the calibration panes.
        """
        muon_monitor = cls(base_path=base_path, notebook=notebook)
        sidebar = muon_monitor.build_sidebar()
        return pn.Row(
            sidebar, pn.Tabs(*muon_monitor.build_muon_panes(widget_widths).values())
        )


def run_dashboard_muon() -> None:
    argparser = argparse.ArgumentParser()
    argparser.add_argument("config_file", type=str)
    argparser.add_argument("-p", "--port", type=int, default=9000)
    argparser.add_argument(
        "-w", "--widget_widths", type=int, default=140, required=False
    )
    args = argparser.parse_args()

    config = read_config(args.config_file)
    muon_panes = MuonMonitoring.display_cal_panes(config.base, args.widget_widths)
    print("Starting Muon Monitoring on port ", args.port)  # noqa: T201
    pn.serve(muon_panes, port=args.port, show=False)



from __future__ import annotations

from legenddashboard.geds.phy.phy_plots import (
    phy_plot_histogram,
    phy_plot_vsTime,
    phy_plots_sc_vals_dict,
    phy_plots_types_dict,
    phy_plots_vals_dict,
    phy_resampled_vals,
    phy_unit_vals,
)

__all__ = [
    "phy_plot_histogram",
    "phy_plot_vsTime",
    "phy_plots_sc_vals_dict",
    "phy_plots_types_dict",
    "phy_plots_vals_dict",
    "phy_resampled_vals",
    "phy_unit_vals",
]

phy_plot_style_dict = {
    "Time": phy_plot_vsTime,
    "Histogram": phy_plot_histogram,
}



from __future__ import annotations

import argparse
import logging
import time

import pandas as pd
import panel as pn
import param
from bokeh.models import ColumnDataSource
from bokeh.models.widgets.tables import BooleanFormatter

import legenddashboard.geds.string_visulization as visu
from legenddashboard.base import Monitoring
from legenddashboard.util import (
    get_characterization,
    get_production,
    logo_path,
    read_config,
    sort_dict,
    sorter,
)

log = logging.getLogger(__name__)

meta_visu_plots_dict = {
    "Usability": visu.plot_visu_usability,
    "Processable": visu.plot_visu_processable,
    "Mass": visu.plot_visu_mass,
    "Depl. Voltage": visu.plot_visu_depletion,
    "Oper. Voltage": visu.plot_visu_operation,
    "Enrichment": visu.plot_visu_enrichment,
}


class GedMonitoring(Monitoring):
    channel = param.Selector(default=0, objects=[0], allow_refs=True, nested_refs=True)
    string = param.ObjectSelector(
        default="01",
        objects=[f"{i + 1:02}" for i in range(11)],
        allow_refs=True,
        nested_refs=True,
    )
    # general selectors
    sort_by = param.ObjectSelector(
        default=next(iter(sort_dict)),
        objects=list(sort_dict),
        allow_refs=True,
        nested_refs=True,
    )

    meta_visu_plots = param.ObjectSelector(
        default=next(iter(meta_visu_plots_dict)), objects=list(meta_visu_plots_dict)
    )
    meta_visu_plots_dict = param.Dict(meta_visu_plots_dict)
    meta_df = pd.DataFrame()
    meta_visu_source = ColumnDataSource({})
    meta_visu_xlabels = param.Dict({})
    meta_visu_chan_dict = param.Dict({})
    meta_visu_channel_map = param.Dict({})

    chan_dict = param.Dict({})
    strings_dict = param.Dict({})
    channel_map = param.Dict({})

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.param.watch(
            self.get_run_and_channel,
            ["period", "run", "channel"],
            precedence=2,
            queued=True,
        )
        self.param.watch(
            self.update_strings, ["period", "run", "sort_by"], precedence=1, queued=True
        )
        self.param.watch(
            self._get_metadata, ["period", "run"], precedence=1, queued=True
        )

    def get_run_and_channel(self, event=None):  # noqa: ARG002
        try:
            start_time = time.time()
            ret = pn.pane.Markdown(
                f"### {self.run_dict[self.run]['experiment']}-{self.period}-{self.run} | Cal. Details | Channel {self.channel}"
            )
            log.debug(
                "Time to get run and channel:", extra={"time": time.time() - start_time}
            )
        except BaseException:
            ret = pn.pane.Markdown("###")
        return ret

    def update_strings(self, event=None):  # noqa: ARG002
        start_time = time.time()
        strings_dict, self.chan_dict, self.channel_map = sorter(
            self.base_path,
            self.run_dict[self.run]["timestamp"],
            key=self.sort_by,
            sort_dets_obj=self.sort_obj,
        )

        self.param["string"].objects = list(strings_dict)
        self.string = f"{next(iter(strings_dict))}"
        self.strings_dict = strings_dict
        log.debug("Time to update strings:", extra={"time": time.time() - start_time})

    def _get_metadata(self, event=None):  # noqa: ARG002
        start_time = time.time()
        try:
            chan_dict, channel_map = self.chan_dict, self.channel_map

            df_chan_dict = pd.DataFrame.from_dict(chan_dict).T
            df_chan_dict.index.name = "name"
            df_chan_dict = df_chan_dict.reset_index()

            df_channel_map = pd.DataFrame.from_dict(channel_map).T
            df_channel_map = df_channel_map[df_channel_map["system"] == "geds"]

            df_out = df_channel_map.merge(df_chan_dict, left_on="name", right_on="name")
            df_out = df_out.reset_index().set_index("name")[
                [
                    "processable",
                    "usability",
                    "daq",
                    "location",
                    "voltage",
                    "electronics",
                    "characterization",
                    "production",
                    "type",
                ]
            ]
            df_out["daq"] = df_out["daq"].map(
                lambda x: "Crate: {}, Card: {}".format(x["crate"], x["card"]["id"])
            )
            df_out["location"] = df_out["location"].map(
                lambda x: "String: {:>02d}, Pos.: {:>02d}".format(
                    x["string"], x["position"]
                )
            )
            df_out["voltage"] = df_out["voltage"].map(
                lambda x: "Card: {:>02d}, Ch.: {:>02d}".format(
                    x["card"]["id"], x["channel"]
                )
            )
            df_out["electronics"] = df_out["electronics"].map(
                lambda x: "CC4: {}, Ch.: {:>02d}".format(
                    x["cc4"]["id"], x["cc4"]["channel"]
                )
            )
            df_out["usability"] = df_out["usability"].map(lambda x: x == "on")
            # df_out['processable'] =  df_out['processable'].map(lambda x: True if x == 'True' else False)
            df_out["Depl. Vol. (kV)"] = (
                df_out["characterization"].map(
                    lambda x: get_characterization(x, "depletion_voltage_in_V")
                )
                / 1000
            )
            df_out["Oper. Vol. (kV)"] = (
                df_out["characterization"].map(
                    lambda x: get_characterization(x, "recommended_voltage_in_V")
                )
                / 1000
            )
            df_out["Manufacturer"] = df_out["production"].map(
                lambda x: get_production(x, "manufacturer")
            )
            df_out["Mass (kg)"] = (
                df_out["production"].map(lambda x: get_production(x, "mass_in_g"))
                / 1000
            )
            df_out["Order"] = df_out["production"].map(
                lambda x: get_production(x, "order")
            )
            df_out["Crystal"] = df_out["production"].map(
                lambda x: get_production(x, "crystal")
            )
            df_out["Slice"] = df_out["production"].map(
                lambda x: get_production(x, "slice")
            )
            df_out["Enrichment (%)"] = (
                df_out["production"].map(
                    lambda x: get_production(x, "enrichment")["val"]
                )
                * 100
            )
            df_out["Delivery"] = df_out["production"].map(
                lambda x: get_production(x, "delivered")
            )
            df_out = (
                df_out.reset_index()
                .rename(
                    {
                        "name": "Det. Name",
                        "processable": "Proc.",
                        "usability": "Usabl.",
                        "daq": "FC card",
                        "location": "Det. Location",
                        "voltage": "HV",
                        "electronics": "Electronics",
                        "type": "Type",
                    },
                    axis=1,
                )
                .set_index("Det. Name")
            )
            df_out = df_out.drop(["characterization", "production"], axis=1)
            df_out = df_out.astype({"Proc.": "bool", "Usabl.": "bool"})
            self.meta_df = df_out

        except KeyError:
            pass
        log.debug("Time to get metadata:", extra={"time": time.time() - start_time})

    @param.depends("run")
    def view_meta(self):
        start_time = time.time()
        ret = pn.widgets.Tabulator(
            self.meta_df,
            formatters={"Proc.": BooleanFormatter(), "Usabl.": BooleanFormatter()},
            frozen_columns=[0],
        )
        log.debug("Time to get meta:", extra={"time": time.time() - start_time})
        return ret

    @param.depends("run", "meta_visu_plots")
    def view_meta_visu(self):
        start_time = time.time()
        strings_dict, meta_visu_chan_dict, meta_visu_channel_map = sorter(
            self.base_path,
            self.run_dict[self.run]["timestamp"],
            key="String",
            sort_dets_obj=self.sort_obj,
        )
        meta_visu_source, meta_visu_xlabels = visu.get_plot_source_and_xlabels(
            meta_visu_chan_dict, meta_visu_channel_map, strings_dict
        )
        figure = None
        figure = self.meta_visu_plots_dict[self.meta_visu_plots](
            meta_visu_source,
            meta_visu_chan_dict,
            meta_visu_channel_map,
            meta_visu_xlabels,
        )
        log.debug("Time to get meta visu:", extra={"time": time.time() - start_time})
        return figure

    def build_sidebar(self, sidebar_instance=None):
        string_param = pn.widgets.MenuButton(
            name=f"{self.string}",
            button_type="primary",
            sizing_mode="stretch_width",
            items=self.param.string.objects,
        )

        def update_string(event):
            self.string = event.new
            string_param.name = f"{self.string}"

        string_param.on_click(update_string)

        sort_by_param = pn.widgets.MenuButton(
            name=f"Sorted by {self.sort_by}",
            button_type="primary",
            sizing_mode="stretch_width",
            items=self.param.sort_by.objects,
        )

        def update_sort_by(event):
            self.sort_by = event.new
            string_param.items = self.param.string.objects
            string_param.name = f"{self.string}"
            sort_by_param.name = f"Sorted by {self.sort_by}"

        sort_by_param.on_click(update_sort_by)
        # sort_by_param    = pn.Param(cal.param, widgets={'sort_by': {'widget_type': pn.widgets.Select, 'width': 100}}, parameters=['sort_by'], show_labels=False, show_name=False, design=Bootstrap)

        if sidebar_instance is not None:
            sidebar = sidebar_instance
        else:
            sidebar = super().build_sidebar()
        sidebar.append(
            pn.pane.SVG(
                logo_path / "Sort_by.svg",
                height=25,
            )
        )
        sidebar.append(sort_by_param)
        sidebar.append(pn.pane.SVG(logo_path / "String.svg", height=25))
        sidebar.append(string_param)
        return sidebar

    def build_meta_pane(self, widget_widths=130):
        self.update_strings(None)
        self._get_metadata(None)
        meta_param_currentValue = pn.pane.Markdown(f"## {self.meta_visu_plots}")
        meta_param = pn.widgets.MenuButton(
            name="Visualization",
            button_type="primary",
            width=widget_widths,
            items=self.param.meta_visu_plots.objects,
        )

        def update_meta_plots(event):
            self.meta_visu_plots = event.new
            meta_param_currentValue.object = f"## {event.new}"

        meta_param.on_click(update_meta_plots)

        return pn.Column(
            pn.Row(
                pn.pane.SVG(
                    logo_path / "Metadata.svg",
                    height=25,
                ),
                meta_param,
            ),
            pn.Row("## Current Plot:", meta_param_currentValue),
            pn.Tabs(
                (
                    "Detector Visu",
                    pn.pane.Bokeh(self.view_meta_visu, sizing_mode="scale_both"),
                ),
                ("Detector Map", self.view_meta),
                sizing_mode="scale_both",
            ),
            name="MetaData",
            sizing_mode="scale_both",
        )

    @classmethod
    def build_ged_base(
        cls,
        path,
        notebook=False,
        widget_widths=130,
        monitoring_instance=None,
        sidebar_instance=None,
    ):
        if monitoring_instance is not None:
            ged_monitoring = cls(
                base_path=path,
                notebook=notebook,
                run_dict=monitoring_instance.params.run_dict,
                period=monitoring_instance.params.period,
                prod_config=monitoring_instance.params.prod_config,
            )
        else:
            ged_monitoring = cls(
                base_path=path,
                notebook=notebook,
            )
        sidebar = ged_monitoring.build_sidebar(sidebar_instance)
        meta_pane = ged_monitoring.build_meta_pane(widget_widths)
        return sidebar, meta_pane, ged_monitoring

    @classmethod
    def display_meta(
        cls,
        path,
        notebook=False,
        widget_widths=130,
    ):
        sidebar, meta_pane, _ = cls.build_ged_base(
            path=path,
            notebook=notebook,
            widget_widths=widget_widths,
        )
        return pn.Row(
            sidebar,
            meta_pane,
            sizing_mode="scale_both",
        )


def run_dashboard_meta() -> None:
    argparser = argparse.ArgumentParser()
    argparser.add_argument("config_file", type=str)
    argparser.add_argument("-p", "--port", type=int, default=9000)
    argparser.add_argument(
        "-w", "--widget_widths", type=int, default=140, required=False
    )
    args = argparser.parse_args()

    config = read_config(args.config_file)
    meta_panes = GedMonitoring.display_meta(config.base, args.widget_widths)
    print("Starting Meta. Monitoring on port ", args.port)  # noqa: T201
    pn.serve(meta_panes, port=args.port, show=False)



from __future__ import annotations

import argparse
import logging
import time
from pathlib import Path

import holoviews as hv
import numpy as np
import pandas as pd
import panel as pn
import param
from bokeh.plotting import figure

from legenddashboard.util import read_config

log = logging.getLogger(__name__)


class LlamaMonitoring(param.Parameterized):
    llama_path = param.String("")

    def view_llama(self):
        start_time = time.time()
        try:
            llama_data = pd.read_csv(
                self.llama_path + "monivalues.txt",
                sep=r"\s+",
                dtype={"timestamp": np.int64},
                parse_dates=[1],
            )
            llama_data["timestamp"] = pd.to_datetime(
                llama_data["timestamp"], origin="unix", unit="s"
            )
            llama_data = llama_data.rename(
                columns={
                    "#run_no": "#Run",
                    "timestamp": "Timestamp",
                    "triplet_val": "Triplet Lifetime (µs)",
                    "triplet_err": "Error Triplet Lifetime (µs)",
                    "ly_val": "Light Yield",
                    "ly_err": "Error Light Yield",
                }
            )
        except KeyError:
            p = figure(width=1000, height=600)
            p.title.text = "No current Llama data available."
            p.title.align = "center"
            p.title.text_font_size = "25px"
            log.debug(
                "Time to get llama plot:", extra={"time": time.time() - start_time}
            )
            return p
        llama_width, llama_height = 1200, 400
        # add two hours to x values to convert from UTC to CET if values still in UTC
        if llama_data["Timestamp"][0].utcoffset() is None:
            llama_data["Timestamp"] += pd.Timedelta(hours=2)
        triplet_plot = hv.Scatter(
            llama_data, ["Timestamp", "Triplet Lifetime (µs)"], label="Triplet LT"
        )
        triplet_plot_error = hv.ErrorBars(
            llama_data,
            vdims=["Triplet Lifetime (µs)", "Error Triplet Lifetime (µs)"],
            kdims=["Timestamp"],
            label="Triplet LT Error",
        )
        triplet_plot.opts(
            xlabel="Time (CET)",
            ylabel="Triplet Lifetime (µs)",
            tools=["hover"],
            line_width=1.5,
            color="blue",
            width=llama_width,
            height=llama_height,
        )
        triplet_plot_error.opts(
            line_width=0.2, width=llama_width, height=llama_height, show_grid=True
        )

        lightyield_plot = hv.Scatter(
            llama_data, ["Timestamp", "Light Yield"], label="Light Yield"
        )
        lightyield_plot_error = hv.ErrorBars(
            llama_data,
            vdims=["Light Yield", "Error Light Yield"],
            kdims=["Timestamp"],
            label="Light Yield Error",
        )
        lightyield_plot.opts(
            xlabel="Time (CET)",
            ylabel="Light yield (a.u.)",
            tools=["hover"],
            line_width=1.5,
            color="orange",
            width=llama_width,
            height=llama_height,
            show_grid=True,
        )
        lightyield_plot_error.opts(
            line_width=0.2, width=llama_width, height=llama_height
        )

        layout = (
            triplet_plot * triplet_plot_error + lightyield_plot * lightyield_plot_error
        )
        layout.opts(width=llama_width, height=llama_height).cols(1)
        log.debug("Time to get llama plot:", extra={"time": time.time() - start_time})
        return layout

    def get_llama_lastUpdate(self):
        start_time = time.time()
        llama_pathlib = Path(self.llama_path + "monivalues.txt")
        ret = "Last modified: {}".format(
            pd.to_datetime(llama_pathlib.stat().st_mtime, origin="unix", unit="s")
        )
        log.debug(
            "Time to get llama last update:", extra={"time": time.time() - start_time}
        )
        return ret

    def build_llama_pane(self):
        return pn.Column(
            "# Llama Monitoring",
            self.get_llama_lastUpdate,
            pn.panel(self.view_llama, sizing_mode="scale_both"),
            name="Llama",
            sizing_mode="scale_both",
        )

    @classmethod
    def init_llama_pane(cls, llama_path: str) -> pn.Column:
        """
        Initialize the Llama monitoring pane.
        :param llama_path: Path to the Llama data directory.
        :return: A Panel Column containing the Llama monitoring information.
        """
        llama_monitor = cls(llama_path=llama_path)
        return llama_monitor.build_llama_pane()


def run_dashboard_llama() -> None:
    argparser = argparse.ArgumentParser()
    argparser.add_argument("config_file", type=str)
    argparser.add_argument("-p", "--port", type=int, default=9000)
    argparser.add_argument(
        "-w", "--widget_widths", type=int, default=140, required=False
    )
    args = argparser.parse_args()

    config = read_config(args.config_file)
    llama_pane = LlamaMonitoring.init_llama_pane(config.base)
    print("Starting Llama Monitoring on port ", args.port)  # noqa: T201
    pn.serve(llama_pane, port=args.port, show=False)



from __future__ import annotations

import numpy as np
import pandas as pd
from bokeh.models import (
    DatetimeTickFormatter,
    LinearAxis,
    Range1d,
    ZoomInTool,
    ZoomOutTool,
)
from bokeh.palettes import Category20, Turbo256
from bokeh.plotting import figure
from seaborn import color_palette

# physics plots
phy_plots_types_dict = {
    "Pulser Events": "IsPulser",
    "Baseline Events": "IsBsln",
}
phy_plots_vals_dict = {
    "Baseline FPGA": "Baseline",
    "Baseline Mean": "BlMean",
    "Noise": "BlStd",
    "Gain": "Cuspemax",
    "Cal. Gain": "CuspemaxCtcCal",
    "Gain to Pulser Ratio": "Cuspemax_pulser01anaRatio",
    "Gain to Pulser Diff.": "Cuspemax_pulser01anaDiff",
    "Rate": "EventRate",
    "PSD Classifier": "AoeCustom",
}
phy_resampled_vals = [0, 5, 10, 30, 60]
phy_unit_vals = ["Relative", "Absolute"]
phy_plots_sc_vals_dict = {
    "None": False,
    "DAQ Temp. Left 1": "DaqLeft_Temp1",
    "DAQ Temp. Left 2": "DaqLeft_Temp2",
    "DAQ Temp. Right 1": "DaqRight_Temp1",
    "DAQ Temp. Right 2": "DaqRight_Temp2",
    "RREiT": "RREiT",
    "RRNTe": "RRNTe",
    "RRSTe": "RRSTe",
    "ZUL_T_RR": "ZUL_T_RR",
}


def phy_plot_vsTime(
    data_string,
    data_string_mean,
    plot_info,
    plot_type,
    plot_name,
    resample_unit,
    string,
    run,
    period,
    run_dict,
    channel_map,
    abs_unit,
    data_sc,
    sc_param,
):
    # change column names to detector names
    data_string.columns = [
        "{}_val".format(channel_map[ch]["name"]) for ch in data_string.columns
    ]

    # create plot colours
    len_colours = len(data_string.columns)
    colours = color_palette("hls", len_colours).as_hex()

    # add mean values for hover feature
    data_string_mean.columns = [
        channel_map[ch]["name"] for ch in data_string_mean.columns
    ]
    for col in data_string_mean.columns:
        data_string[col] = data_string_mean[col][0]

    # add two hours to x values with if condition
    if data_string.index[0].utcoffset() != pd.Timedelta(
        hours=2
    ):  # only add timedelta if still in UTC
        data_string.index += pd.Timedelta(hours=2)

    p = figure(
        width=1000,
        height=600,
        x_axis_type="datetime",
        tools="pan,box_zoom,ywheel_zoom,hover,reset,save",
        output_backend="webgl",
        active_scroll="ywheel_zoom",
    )
    p.title.text = f"{run_dict['experiment']}-{period}-{run} | Phy. {plot_type} | {plot_name} | {string}"
    p.title.align = "center"
    p.title.text_font_size = "25px"
    p.hover.formatters = {"$x": "datetime", "$snap_y": "printf", "@$name": "printf"}
    p.hover.tooltips = [
        (
            "Time",
            "$x{%F %H:%M:%S CET}",
        ),  # Use the formatted CET time from the DataFrame
        (f"{plot_info.loc['label'][0]} ({plot_info.loc['unit'][0]})", "$snap_y{%0.2f}"),
        (f"Mean {plot_info.loc['label'][0]} ({abs_unit})", "@$name{0.2f}"),
        ("Detector", "$name"),
    ]
    p.hover.mode = "vline"

    level = 1
    zoom_in = ZoomInTool(
        level=level, dimensions="height", factor=0.5
    )  # set specific zoom factor
    zoom_out = ZoomOutTool(level=level, dimensions="height", factor=0.5)
    p.add_tools(zoom_in, zoom_out)
    # p.toolbar.active_drag = None      use this line to activate only hover and ywheel_zoom as active tool

    # plot data
    hover_renderers = []
    if resample_unit == "0min":
        for i, det in enumerate(data_string_mean):
            if "mean" in det:
                continue
            line = p.line(
                "datetime",
                f"{det}_val",
                source=data_string,
                color=colours[i],
                legend_label=det,
                name=det,
                line_width=2.5,
            )
            hover_renderers.append(line)
    else:
        data_string_resampled = data_string.resample(
            resample_unit, origin="start"
        ).mean()

        for i, det in enumerate(data_string_mean):
            if "mean" in det:
                continue
            line = p.line(
                "datetime",
                f"{det}_val",
                source=data_string_resampled,
                color=colours[i],
                legend_label=det,
                name=det,
                line_width=2.5,
            )
            p.line(
                "datetime",
                f"{det}_val",
                source=data_string,
                color=colours[i],
                legend_label=det,
                name=det,
                line_width=2.5,
                alpha=0.2,
            )
            hover_renderers.append(line)

    # draw horizontal line at thresholds from plot info if available
    #     if plot_info.loc["lower_lim_var"][0] != 'None' and plot_info.loc["unit"][0] == "%":
    #         lower_lim_var = Slope(gradient=0, y_intercept=float(plot_info.loc["lower_lim_var"][0]),
    #                 line_color='black', line_dash='dashed', line_width=4)
    #         upper_lim_var = Slope(gradient=0, y_intercept=float(plot_info.loc["upper_lim_var"][0]),
    #                 line_color='black', line_dash='dashed', line_width=4)

    #         p.add_layout(lower_lim_var)
    #         p.add_layout(upper_lim_var)

    # legend setups etc...
    p.legend.location = "bottom_left"
    p.legend.click_policy = "hide"
    p.xaxis.axis_label = f"Time (CET), starting: {data_string.index[0].strftime('%d/%m/%Y %H:%M:%S')}"  # change of string
    p.xaxis.axis_label_text_font_size = "20px"
    p.yaxis.axis_label = f"{plot_info.loc['label'][0]} [{plot_info.loc['unit'][0]}]"
    p.yaxis.axis_label_text_font_size = "20px"
    p.xaxis.formatter = DatetimeTickFormatter(days="%Y/%m/%d")
    p.hover.renderers = hover_renderers

    if plot_info.loc["unit"][0] == "%":
        if plot_info.loc["label"][0] == "Noise":
            p.y_range = Range1d(-150, 150)
        elif (
            plot_info.loc["label"][0] == "FPGA baseline"
            or plot_info.loc["label"][0] == "Mean Baseline"
        ):
            p.y_range = Range1d(-10, 10)
        elif plot_info.loc["label"][0] == "Gain to Pulser Difference":
            p.y_range = Range1d(-4, 4)
        elif plot_info.loc["label"][0] == "Event Rate":
            p.y_range = Range1d(-150, 50)
        elif plot_info.loc["label"][0] == "Custom A/E (A_max / cuspEmax)":
            p.y_range = Range1d(-10, 10)
        else:
            p.y_range = Range1d(-1, 1)
    elif plot_info.loc["label"][0] == "Noise":
        p.y_range = Range1d(-150, 150)

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # SLOW CONTROL DATA
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    if not data_sc.empty:
        y_column2_range = f"{sc_param}_range"
        y_min = float(data_sc.copy()["value"].min()) * (1 - 0.01)
        y_max = float(data_sc.copy()["value"].max()) * (1 + 0.01)
        p.extra_y_ranges = {y_column2_range: Range1d(start=y_min, end=y_max)}

        unit = data_sc["unit"][0]
        p.add_layout(
            LinearAxis(
                y_range_name=y_column2_range,
                axis_label=f"{sc_param} [{unit}]",
                axis_label_text_font_size="20px",
            ),
            "right",
        )

        time = data_sc["tstamp"]
        time = pd.to_datetime(time, origin="unix", utc=True)
        values = data_sc["value"]
        values = pd.to_numeric(values)
        values.index = time

        # we use the same resampling of geds data
        # (use black line to distinguish from geds data)
        if resample_unit == "0min":
            p.line(
                time,
                values,
                legend_label=sc_param,
                y_range_name=y_column2_range,
                color="black",
                line_width=2,
            )
        else:
            binned_data = values.resample(resample_unit).mean()
            p.line(
                time,
                values,
                color="black",
                legend_label=sc_param,
                y_range_name=y_column2_range,
                line_width=2,
                alpha=0.2,
            )
            p.line(
                binned_data.index,
                binned_data.values,
                color="black",
                legend_label=sc_param,
                y_range_name=y_column2_range,
                line_width=2,
            )

    return p


def phy_plot_histogram(
    data_string,
    plot_info,
    plot_type,
    resample_unit,
    string,
    run,
    period,
    run_dict,
    channels,
    channel_map,
):
    p = figure(
        width=1000,
        height=600,
        x_axis_type="datetime",
        tools="pan,box_zoom,ywheel_zoom,hover,reset,save",
        output_backend="webgl",
        active_scroll="ywheel_zoom",
    )
    p.title.text = f"{run_dict['experiment']}-{period}-{run} | Phy. {plot_type} | {plot_info.loc['label'][0]} | {string}"
    p.title.align = "center"
    p.title.text_font_size = "25px"
    p.hover.formatters = {"$x": "printf", "$snap_y": "printf"}
    p.hover.tooltips = [
        (f"{plot_info.loc['label'][0]} ({plot_info.loc['unit'][0]}", "$x{%0.2f}"),
        ("Counts", "$snap_y"),
        ("Detector", "$name"),
    ]

    p.hover.mode = "vline"

    level = 1
    zoom_in = ZoomInTool(
        level=level, dimensions="height", factor=0.5
    )  # set specific zoom factor
    zoom_out = ZoomOutTool(level=level, dimensions="height", factor=0.5)
    p.add_tools(zoom_in, zoom_out)
    # p.toolbar.active_drag = None      use this line to activate only hover and ywheel_zoom as active tool

    len_colours = len(data_string.columns)
    colours = Turbo256[len_colours] if len_colours > 19 else Category20[len_colours]

    for position, data_channel in data_string.groupby("position"):
        # generate histogram
        # needed for cuspEmax because with geant outliers not possible to view normal histo
        hrange = {"keV": [0, 2500]}
        # take full range if not specified
        x_min = (
            hrange[plot_info["unit"]][0]
            if plot_info["unit"] in hrange
            else data_channel[plot_info["parameter"]].min()
        )
        x_max = (
            hrange[plot_info["unit"]][1]
            if plot_info["unit"] in hrange
            else data_channel[plot_info["parameter"]].max()
        )

        # --- bin width
        # bwidth = {"keV": 2.5}  # what to do with binning???
        # bin_width = bwidth[plot_info["unit"]] if plot_info["unit"] in bwidth else None
        # no_bins = int((x_max - x_min) / bin_width) if bin_width else 50
        # counts_ch, bins_ch = np.histogram(data_channel[plot_info["parameter"]], bins=no_bins, range=(x_min, x_max))
        # bins_ch = (bins_ch[:-1] + bins_ch[1:]) / 2

        # --- bin width
        bwidth = {"keV": 2.5}
        bin_width = bwidth.get(plot_info["unit"], 1)

        # Compute number of bins
        if bin_width:
            bin_no = (
                bin_width / 5 if "AoE" not in plot_info["parameter"] else bin_width / 50
            )
            bin_no = bin_no / 2 if "Corrected" in plot_info["parameter"] else bin_no
            bin_no = bin_width if "AoE" not in plot_info["parameter"] else bin_no

            bin_edges = (
                np.arange(x_min, x_max + bin_width, bin_no)
                if plot_info["unit_label"] == "%"
                else np.arange(x_min, x_max + bin_width, bin_no)
            )
        else:
            bin_edges = 50
        counts_ch, bins_ch = np.histogram(
            data_channel[plot_info["parameter"]], bins=bin_edges, range=(x_min, x_max)
        )
        bins_ch = (bins_ch[:-1] + bins_ch[1:]) / 2
        # create plot histo
        histo_df = pd.DataFrame(
            {
                "counts": counts_ch,
                "bins": bins_ch,
                "position": position,
                "cc4_id": data_channel["cc4_id"].unique()[0],
            }
        )
        # plot
        p.line(
            "bins",
            "counts",
            source=histo_df,
            color=colours[position - 1],
            legend_label=f"{data_channel['name'].unique()[0]}",
            name=f"ch {data_channel['channel'].unique()[0]}",
            line_width=2,
        )

    p.legend.location = "bottom_left"
    p.legend.click_policy = "hide"
    p.xaxis.axis_label = f"{plot_info['label']} [{plot_info['unit_label']}]"
    p.xaxis.axis_label_text_font_size = "20px"
    p.yaxis.axis_label = "Counts"
    p.yaxis.axis_label_text_font_size = "20px"

    return p



from __future__ import annotations

import contextlib
import datetime as dtt
import warnings
from datetime import datetime
from pathlib import Path

import colorcet as cc
import numexpr as ne
import numpy as np
import pandas as pd
from bokeh.models import (
    Band,
    ColumnDataSource,
    Label,
    Range1d,
    Span,
    ZoomInTool,
    ZoomOutTool,
)
from bokeh.plotting import figure
from dbetto import Props
from scipy.optimize import minimize

from legenddashboard.util import sorter


def plot_energy(path, run_dict, det, plot, colour, period, cache_data=None):
    cals = []
    times = []
    qbb_adc = None
    for run in run_dict:
        if cache_data is not None and run in cache_data["hit"]:
            hit_pars_dict = cache_data["hit"][run]
        else:
            prod_config = Path(path) / "dataflow-config.yaml"
            prod_config = Props.read_from(prod_config, subst_pathvar=True)
            hit_pars_file_path = (
                Path(prod_config["paths"]["par_hit"]) / f"cal/{period}/{run}"
            )

            hit_pars_path = (
                Path(hit_pars_file_path)
                / f'{run_dict[run]["experiment"]}-{period}-{run}-cal-{run_dict[run]["timestamp"]}-par_hit.yaml'
            )

            hit_pars_dict = Props.read_from(hit_pars_path)
            cache_data["hit"][run] = hit_pars_dict

        try:
            hit_dict = hit_pars_dict[det]["pars"]["operations"]["cuspEmax_ctc_cal"]
            if qbb_adc is None:

                def find_qbb_adc(val):
                    return ne.evaluate(
                        "abs(" + "cuspEmax_ctc*a +b" + "-2039)",
                        local_dict=dict({"cuspEmax_ctc": val}, a=0.1, b=0),
                    )

                qbb_adc = minimize(find_qbb_adc, 20000)["x"][0]
            out_data = ne.evaluate(
                f"{hit_dict['expression']}",
                local_dict=dict({"cuspEmax_ctc": qbb_adc}, **hit_dict["parameters"]),
            )
            cals.append(out_data)
            times.append(run_dict[run]["timestamp"])
        except KeyError:
            pass
    if len(cals) > 0:
        cals = np.array(cals)
        plot.step(
            [(datetime.strptime(value, "%Y%m%dT%H%M%SZ")) for value in times],
            (cals - cals[0]),
            legend_label=det,
            mode="after",
            line_width=2,
            line_color=colour,
        )

        plot.scatter(
            [(datetime.strptime(value, "%Y%m%dT%H%M%SZ")) for value in times],
            (cals - cals[0]),
            legend_label=det,
            fill_color="white",
            size=8,
            color=colour,
        )

    return plot


def plot_energy_res(
    path, run_dict, det, plot, colour, period, at="Qbb", cache_data=None
):
    reses = []
    times = []
    for run in run_dict:
        if cache_data is not None and run in cache_data["hit"]:
            hit_pars_dict = cache_data["hit"][run]
        else:
            prod_config = Path(path) / "dataflow-config.yaml"
            prod_config = Props.read_from(prod_config, subst_pathvar=True)

            hit_pars_file_path = (
                Path(prod_config["paths"]["par_hit"]) / f"cal/{period}/{run}"
            )

            hit_pars_path = (
                hit_pars_file_path
                / f'{run_dict[run]["experiment"]}-{period}-{run}-cal-{run_dict[run]["timestamp"]}-par_hit.yaml'
            )

            hit_pars_dict = Props.read_from(hit_pars_path)
            cache_data["hit"][run] = hit_pars_dict
        try:
            if at == "Qbb":
                reses.append(
                    hit_pars_dict[det]["results"]["ecal"]["cuspEmax_ctc_cal"][
                        "eres_linear"
                    ]["Qbb_fwhm_in_kev"]
                )
            elif at == "2.6":
                reses.append(
                    hit_pars_dict[det]["results"]["ecal"]["cuspEmax_ctc_cal"][
                        "pk_fits"
                    ][2614.511]["fwhm_in_kev"]
                )
            times.append(run_dict[run]["timestamp"])
        except KeyError:
            pass

    if len(reses) > 0:
        plot.step(
            [(datetime.strptime(value, "%Y%m%dT%H%M%SZ")) for value in times],
            (reses),
            legend_label=det,
            mode="after",
            line_width=2,
            line_color=colour,
        )

        plot.scatter(
            [(datetime.strptime(value, "%Y%m%dT%H%M%SZ")) for value in times],
            (reses),
            legend_label=det,
            fill_color="white",
            size=8,
            color=colour,
        )

    return plot


def plot_energy_res_Qbb(path, run_dict, det, plot, colour, period, cache_data=None):
    return plot_energy_res(
        path, run_dict, det, plot, colour, period, at="Qbb", cache_data=cache_data
    )


def plot_energy_res_2614(path, run_dict, det, plot, colour, period, cache_data=None):
    return plot_energy_res(
        path, run_dict, det, plot, colour, period, at="2.6", cache_data=cache_data
    )


def plot_aoe_mean(path, run_dict, det, plot, colour, period, cache_data=None):
    means = []
    mean_errs = []
    reses = []
    res_errs = []
    times = []
    for run in run_dict:
        if cache_data is not None and run in cache_data["hit"]:
            hit_pars_dict = cache_data["hit"][run]
        else:
            prod_config = Path(path) / "dataflow-config.yaml"
            prod_config = Props.read_from(prod_config, subst_pathvar=True)
            hit_pars_file_path = (
                Path(prod_config["paths"]["par_hit"]) / f"cal/{period}/{run}"
            )

            hit_pars_path = (
                hit_pars_file_path
                / f'{run_dict[run]["experiment"]}-{period}-{run}-cal-{run_dict[run]["timestamp"]}-par_hit.yaml'
            )

            hit_pars_dict = Props.read_from(hit_pars_path)
            cache_data["hit"][run] = hit_pars_dict
        try:
            det_dict = hit_pars_dict[det]

            means.append(det_dict["results"]["aoe"]["1000-1300keV"][0]["mean"])
            mean_errs.append(det_dict["results"]["aoe"]["1000-1300keV"][0]["mean_err"])
            reses.append(det_dict["results"]["aoe"]["1000-1300keV"][0]["res"])
            res_errs.append(det_dict["results"]["aoe"]["1000-1300keV"][0]["res_err"])
            times.append(run_dict[run]["timestamp"])
        except KeyError:
            continue
    means = np.array(means)
    reses = np.array(reses)
    plot.step(
        [(datetime.strptime(value, "%Y%m%dT%H%M%SZ")) for value in times],
        100 * (means - means[0]) / reses,
        legend_label=det,
        mode="after",
        line_width=2,
        line_color=colour,
    )

    plot.scatter(
        [(datetime.strptime(value, "%Y%m%dT%H%M%SZ")) for value in times],
        100 * (means - means[0]) / reses,
        legend_label=det,
        fill_color="white",
        size=8,
        color=colour,
    )

    mean_df = pd.DataFrame(
        {
            "x": [(datetime.strptime(value, "%Y%m%dT%H%M%SZ")) for value in times],
            "lower": [-40 for value in times],
            "upper": [40 for value in times],
        }
    )

    source = ColumnDataSource(mean_df)
    band = Band(
        base="x",
        lower="lower",
        upper="upper",
        source=source,
        fill_alpha=0.01,
        fill_color="yellow",
    )
    plot.add_layout(band)

    mean_df2 = pd.DataFrame(
        {
            "x": [(datetime.strptime(value, "%Y%m%dT%H%M%SZ")) for value in times],
            "lower": [-20 for value in times],
            "upper": [20 for value in times],
        }
    )

    source2 = ColumnDataSource(mean_df2)
    band2 = Band(
        base="x",
        lower="lower",
        upper="upper",
        source=source2,
        fill_alpha=0.02,
        fill_color="green",
    )
    plot.add_layout(band2)
    plot.y_range = Range1d(-100, 100)

    return plot


def plot_aoe_sig(path, run_dict, det, plot, colour, period, cache_data=None):
    means = []
    mean_errs = []
    reses = []
    res_errs = []
    times = []
    for run in run_dict:
        if cache_data is not None and run in cache_data["hit"]:
            hit_pars_dict = cache_data["hit"][run]
        else:
            prod_config = Path(path) / "dataflow-config.yaml"
            prod_config = Props.read_from(prod_config, subst_pathvar=True)
            hit_pars_file_path = (
                Path(prod_config["paths"]["par_hit"]) / f"cal/{period}/{run}"
            )

            hit_pars_path = (
                hit_pars_file_path
                / f'{run_dict[run]["experiment"]}-{period}-{run}-cal-{run_dict[run]["timestamp"]}-par_hit.yaml'
            )

            hit_pars_dict = Props.read_from(hit_pars_path)
            cache_data["hit"][run] = hit_pars_dict

        try:
            means.append(
                hit_pars_dict[det]["results"]["aoe"]["1000-1300keV"][0]["mean"]
            )
            mean_errs.append(
                hit_pars_dict[det]["results"]["aoe"]["1000-1300keV"][0]["mean_err"]
            )
            reses.append(hit_pars_dict[det]["results"]["aoe"]["1000-1300keV"][0]["res"])
            res_errs.append(
                hit_pars_dict[det]["results"]["aoe"]["1000-1300keV"][0]["res_err"]
            )
            times.append(run_dict[run]["timestamp"])
        except KeyError:
            pass
    means = np.array(means)
    reses = np.array(reses)
    plot.step(
        [(datetime.strptime(value, "%Y%m%dT%H%M%SZ")) for value in times],
        (reses),
        legend_label=det,
        mode="after",
        line_width=2,
        line_color=colour,
    )

    plot.scatter(
        [(datetime.strptime(value, "%Y%m%dT%H%M%SZ")) for value in times],
        (reses),
        legend_label=det,
        fill_color="white",
        size=8,
        color=colour,
    )

    return plot


def plot_aoe_cut(path, run_dict, det, plot, colour, period, cache_data=None):
    cuts = []
    times = []
    for run in run_dict:
        if cache_data is not None and run in cache_data["hit"]:
            hit_pars_dict = cache_data["hit"][run]
        else:
            prod_config = Path(path) / "dataflow-config.yaml"
            prod_config = Props.read_from(prod_config, subst_pathvar=True)
            hit_pars_file_path = (
                Path(prod_config["paths"]["par_hit"]) / f"cal/{period}/{run}"
            )

            hit_pars_path = (
                hit_pars_file_path
                / f'{run_dict[run]["experiment"]}-{period}-{run}-cal-{run_dict[run]["timestamp"]}-par_hit.yaml'
            )

            hit_pars_dict = Props.read_from(hit_pars_path)
            cache_data["hit"][run] = hit_pars_dict

        with contextlib.suppress(KeyError):
            cuts.append(hit_pars_dict[det]["results"]["aoe"]["low_cut"])
            times.append(run_dict[run]["timestamp"])

    cuts = np.array(cuts)
    plot.step(
        [(datetime.strptime(value, "%Y%m%dT%H%M%SZ")) for value in times],
        (cuts),
        legend_label=det,
        mode="after",
        line_width=2,
        line_color=colour,
    )

    plot.scatter(
        [(datetime.strptime(value, "%Y%m%dT%H%M%SZ")) for value in times],
        (cuts),
        legend_label=det,
        fill_color="white",
        size=8,
        color=colour,
    )

    return plot


def plot_tau(path, run_dict, det, plot, colour, period, cache_data=None):
    values = []
    times = []
    for run in run_dict:
        if cache_data is not None and run in cache_data["dsp"]:
            dsp_pars_dict = cache_data["dsp"][run]
        else:
            prod_config = Path(path) / "dataflow-config.yaml"
            prod_config = Props.read_from(prod_config, subst_pathvar=True)
            dsp_pars_file_path = (
                Path(prod_config["paths"]["par_dsp"]) / f"cal/{period}/{run}"
            )

            dsp_pars_path = (
                dsp_pars_file_path
                / f'{run_dict[run]["experiment"]}-{period}-{run}-cal-{run_dict[run]["timestamp"]}-par_dsp.yaml'
            )

            dsp_pars_dict = Props.read_from(dsp_pars_path)
            cache_data["dsp"][run] = dsp_pars_dict

        with contextlib.suppress(KeyError):
            values.append(float(dsp_pars_dict[det]["pz"]["tau1"][:-3]))
            times.append(run_dict[run]["timestamp"])

    values = np.array(values)
    plot.step(
        [(datetime.strptime(value, "%Y%m%dT%H%M%SZ")) for value in times],
        100 * (values - values[0]) / values[0],
        legend_label=det,
        mode="after",
        line_width=2,
        line_color=colour,
    )

    plot.scatter(
        [(datetime.strptime(value, "%Y%m%dT%H%M%SZ")) for value in times],
        100 * (values - values[0]) / values[0],
        legend_label=det,
        fill_color="white",
        size=8,
        color=colour,
    )

    return plot


def plot_ctc_const(path, run_dict, det, plot, colour, period, cache_data=None):
    values = []
    times = []
    for run in run_dict:
        if cache_data is not None and run in cache_data["dsp"]:
            dsp_pars_dict = cache_data["dsp"][run]
        else:
            prod_config = Path(path) / "dataflow-config.yaml"
            prod_config = Props.read_from(prod_config, subst_pathvar=True)
            dsp_pars_file_path = (
                Path(prod_config["paths"]["par_dsp"]) / f"cal/{period}/{run}"
            )
            dsp_pars_path = (
                dsp_pars_file_path
                / f'{run_dict[run]["experiment"]}-{period}-{run}-cal-{run_dict[run]["timestamp"]}-par_dsp.yaml'
            )

            dsp_pars_dict = Props.read_from(dsp_pars_path)
            cache_data["dsp"][run] = dsp_pars_dict

        with contextlib.suppress(KeyError):
            values.append(
                dsp_pars_dict[det]["ctc_params"]["cuspEmax_ctc"]["parameters"]["a"]
            )
            times.append(run_dict[run]["timestamp"])

    values = np.array(values)
    plot.step(
        [(datetime.strptime(value, "%Y%m%dT%H%M%SZ")) for value in times],
        (values),
        legend_label=det,
        mode="after",
        line_width=2,
        line_color=colour,
    )

    plot.scatter(
        [(datetime.strptime(value, "%Y%m%dT%H%M%SZ")) for value in times],
        (values),
        legend_label=det,
        fill_color="white",
        size=8,
        color=colour,
    )

    return plot


def plot_tracking(
    run_dict,
    path,
    plot_func,
    string,
    period,
    plot_type,
    key="String",
    cache_data=None,
    sort_dets_obj=None,
):
    strings_dict, _, chmap = sorter(
        path,
        run_dict[next(iter(run_dict))]["timestamp"],
        key=key,
        sort_dets_obj=sort_dets_obj,
    )
    string_dets = {}
    for stri in strings_dict:
        dets = []
        for chan in strings_dict[stri]:
            dets.append(chmap[chan]["name"])
        string_dets[stri] = dets

    p = figure(
        width=1000,
        height=400,
        x_axis_type="datetime",
        tools="pan, box_zoom, ywheel_zoom, hover,reset,save",
        active_scroll="ywheel_zoom",
    )
    p.title.text = f"{run_dict[next(iter(run_dict))]['experiment']}-{period} | Cal. Tracking | {plot_type} | {string}"
    p.title.align = "center"
    p.title.text_font_size = "15px"

    level = 1
    zoom_in = ZoomInTool(
        level=level, dimensions="height", factor=0.5
    )  # set specific zoom factor
    zoom_out = ZoomOutTool(level=level, dimensions="height", factor=0.5)
    p.add_tools(zoom_in, zoom_out)
    # p.toolbar.active_drag = None      use this line to activate only hover and ywheel_zoom as active tool

    colours = cc.palette["glasbey_category10"][:100]

    if cache_data is None or len(cache_data) == 0:
        cache_data = {"hit": {}, "dsp": {}}

    for i, det in enumerate(string_dets[string]):
        # with contextlib.suppress(KeyError):
        plot_func(path, run_dict, det, p, colours[i], period, cache_data)

    for run in run_dict:
        sp = Span(
            location=datetime.strptime(run_dict[run]["timestamp"], "%Y%m%dT%H%M%SZ"),
            dimension="height",
            line_color="black",
            line_width=1.5,
        )
        p.add_layout(sp)

        label = Label(
            x=datetime.strptime(run_dict[run]["timestamp"], "%Y%m%dT%H%M%SZ")
            + dtt.timedelta(minutes=200),
            y=0,
            text=run,
        )

        p.add_layout(label)

    p.xaxis.axis_label = "Time"
    p.xaxis.axis_label_text_font_size = "20px"
    p.yaxis.axis_label_text_font_size = "20px"

    if plot_func == plot_energy:
        p.yaxis.axis_label = "Shift of Qbb in keV "
    elif plot_func == plot_energy_res_Qbb:
        p.yaxis.axis_label = "FWHM at Qbb"
    elif plot_func == plot_energy_res_2614:
        p.yaxis.axis_label = "FWHM of 2.6MeV peak"
    elif plot_func == plot_aoe_mean:
        p.yaxis.axis_label = "% Shift of A/E mean"
    elif plot_func == plot_aoe_sig:
        p.yaxis.axis_label = "Shift of A/E Resolution"
    elif plot_func == plot_aoe_sig:
        p.yaxis.axis_label = "Shift of A/E Low Cut"
    elif plot_func == plot_tau:
        p.yaxis.axis_label = "% Shift PZ const"
    elif plot_func == plot_ctc_const:
        p.yaxis.axis_label = "Shift CT constant"

    p.legend.location = "top_left"
    p.legend.click_policy = "hide"

    return p


def plot_energy_residuals_period(
    run_dict,
    path,
    period,
    key="String",
    download=False,
    cache_data=None,
    sort_dets_obj=None,
):
    strings, _, _ = sorter(
        path,
        run_dict[next(iter(run_dict))]["timestamp"],
        key=key,
        sort_dets_obj=sort_dets_obj,
    )

    peaks = [2614.511, 583.191, 2103.511]

    res = {}

    if cache_data is None or len(cache_data) == 0:
        cache_data = {"hit": {}, "dsp": {}}

    for stri in strings:
        res[stri] = {peak: [] for peak in peaks}
        for detector in strings[stri]:
            res[detector] = {peak: [] for peak in peaks}

    for run in run_dict:
        if cache_data is not None and run in cache_data["hit"]:
            hit_pars_dict = cache_data["hit"][run]
        else:
            prod_config = Path(path) / "dataflow-config.yaml"
            prod_config = Props.read_from(prod_config, subst_pathvar=True)
            hit_pars_file_path = (
                Path(prod_config["paths"]["par_hit"]) / f"cal/{period}/{run}"
            )

            hit_pars_path = (
                hit_pars_file_path
                / f'{run_dict[run]["experiment"]}-{period}-{run}-cal-{run_dict[run]["timestamp"]}-par_hit.yaml'
            )

            hit_pars_dict = Props.read_from(hit_pars_path)
            cache_data["hit"][run] = hit_pars_dict

        for peak in peaks:
            for stri in strings:
                res[stri][peak].append(np.nan)
                for detector in strings[stri]:
                    try:
                        hit_dict = hit_pars_dict[detector]["pars"]["operations"][
                            "cuspEmax_ctc_cal"
                        ]
                        res_dict = hit_pars_dict[detector]["results"]["ecal"][
                            "cuspEmax_ctc_cal"
                        ]
                        out_data = (
                            ne.evaluate(
                                f"{hit_dict['expression']}",
                                local_dict=dict(
                                    {
                                        "cuspEmax_ctc": res_dict["pk_fits"][peak][
                                            "parameters"
                                        ]["mu"]
                                    },
                                    **hit_dict["parameters"],
                                ),
                            )
                            - peak
                        )
                        res[detector][peak].append(out_data)
                    except KeyError:
                        continue

    # p = figure(width=1400, height=600, tools="pan,wheel_zoom,box_zoom,xzoom_in,xzoom_out,hover,reset,save")
    p = figure(
        width=1400,
        height=600,
        tools="pan, box_zoom, ywheel_zoom, hover,reset,save",
        active_scroll="ywheel_zoom",
    )
    p.title.text = f"{run_dict[next(iter(run_dict))]['experiment']}-{period} | Cal. | Energy Residuals"
    p.title.align = "center"
    p.title.text_font_size = "25px"

    level = 1
    zoom_in = ZoomInTool(
        level=level, dimensions="height", factor=0.5
    )  # set specific zoom factor
    zoom_out = ZoomOutTool(level=level, dimensions="height", factor=0.5)
    p.add_tools(zoom_in, zoom_out)
    # p.toolbar.active_drag = None      use this line to activate only hover and ywheel_zoom as active tool

    label_res = [r if "String" not in r else "" for r in list(res)]

    df_plot = pd.DataFrame()
    df_plot["label_res"] = label_res

    for peak in peaks:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)
            x_plot, y_plot, y_min, y_max = (
                np.arange(1, len(list(res)) + 1, 1),
                [np.nanmean(res[det][peak]) for det in res],
                [
                    np.nanmin(res[det][peak])
                    if len(np.array(res[det][peak])[~np.isnan(res[det][peak])]) > 0
                    else np.nan
                    for det in res
                ],
                [
                    np.nanmax(res[det][peak])
                    if len(np.array(res[det][peak])[~np.isnan(res[det][peak])]) > 0
                    else np.nan
                    for det in res
                ],
            )

        err_xs = []
        err_ys = []

        for x, y, yerr_low, yerr_hi in zip(x_plot, y_plot, y_min, y_max, strict=False):
            err_xs.append((x, x))
            err_ys.append((np.nan_to_num(y - yerr_low), np.nan_to_num(y + yerr_hi)))

        df_plot[f"x_{int(peak)}"] = np.nan_to_num(x_plot)
        df_plot[f"y_{int(peak)}"] = np.nan_to_num(y_plot)
        df_plot[f"y_{int(peak)}_min"] = np.nan_to_num(y_min)
        df_plot[f"y_{int(peak)}_max"] = np.nan_to_num(y_max)
        df_plot[f"err_xs_{int(peak)}"] = err_xs
        df_plot[f"err_ys_{int(peak)}"] = err_ys

    if download:
        return (
            df_plot,
            f"{run_dict[next(iter(run_dict))]['experiment']}-{period}-energy_residuals.csv",
        )

    for peak, peak_color in zip(peaks, ["blue", "green", "red"], strict=False):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)
            if peak == peaks[0]:
                hover_renderer = p.scatter(
                    x=f"x_{int(peak)}",
                    y=f"y_{int(peak)}",
                    source=df_plot,
                    color=peak_color,
                    size=7,
                    line_alpha=0,
                    legend_label=f"{peak} Average: {np.nanmean([np.nanmean(res[det][peak]) for det in res]):.2f}keV",
                    name=f"{peak} Average: {np.nanmean([np.nanmean(res[det][peak]) for det in res]):.2f}keV",
                )
            else:
                p.scatter(
                    x=f"x_{int(peak)}",
                    y=f"y_{int(peak)}",
                    source=df_plot,
                    color=peak_color,
                    size=7,
                    line_alpha=0,
                    legend_label=f"{peak} Average: {np.nanmean([np.nanmean(res[det][peak]) for det in res]):.2f}keV",
                    name=f"{peak} Average: {np.nanmean([np.nanmean(res[det][peak]) for det in res]):.2f}keV",
                )
            band = Band(
                base=f"x_{int(peak)}",
                lower=f"y_{int(peak)}_min",
                upper=f"y_{int(peak)}_max",
                source=ColumnDataSource(df_plot),
                fill_alpha=0.1,
                fill_color=peak_color,
            )
            p.add_layout(band)

    p.legend.location = "bottom_right"
    p.legend.click_policy = "hide"
    p.xaxis.axis_label = "detector"
    p.xaxis.axis_label_text_font_size = "20px"
    p.yaxis.axis_label = "peak residuals (keV)"
    p.title.text = f"{run_dict[next(iter(run_dict))]['experiment']}-{period} | Cal. | Energy residuals"
    p.yaxis.axis_label_text_font_size = "20px"

    p.xaxis.major_label_orientation = np.pi / 2
    p.xaxis.ticker = np.arange(1, len(list(res)) + 1, 1)
    p.xaxis.major_label_overrides = {
        i: label_res[i - 1] for i in range(1, len(label_res) + 1, 1)
    }
    p.xaxis.major_label_text_font_style = "bold"

    for stri in strings:
        loc = np.where(np.array(list(res)) == stri)[0][0]
        string_span = Span(
            location=loc + 1, dimension="height", line_color="black", line_width=3
        )
        string_span_label = Label(
            x=loc + 1.5, y=1.1, text=stri, text_font_size="10pt", text_color="blue"
        )
        p.add_layout(string_span_label)
        p.add_layout(string_span)

    p.hover.tooltips = [
        ("Detector", "@label_res"),
        ("2614", "av: @y_2614{0.00} min: @y_2614_min{0.00} max: @y_2614_max{0.00} keV"),
        ("SEP", "av: @y_2103{0.00} min: @y_2103_min{0.00} max: @y_2103_max{0.00} keV"),
        ("583", "av: @y_583{0.00} min: @y_583_min{0.00} max: @y_583_max{0.00} keV"),
    ]
    p.hover.mode = "vline"
    p.hover.renderers = [hover_renderer]

    return p
